
import asyncio

from PyQt5.QtWidgets import QWidget, QStackedWidget, QStyle
from PyQt5.QtCore import (QUrl, Qt, pyqtSignal, QRegularExpression,
        QRegularExpressionMatchIterator)
from PyQt5.QtGui import QSyntaxHighlighter, QTextCharFormat, QFont

from galacteek.ipfs.cidhelpers import cidValid
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs import ipfssearch

from . import ui_ipfssearchw, ui_ipfssearchwresults
from . import ipfsview
from .helpers import *
from .hashmarks import *
from .widgets import *
from .i18n import *

def iResultsInfo(count, maxScore):
    return QCoreApplication.translate('IPFSSearchResults',
        'Results count: <b>{0}</b> (max score: <b>{1}</b>)').format(
                count, maxScore)

def iNoResults():
    return QCoreApplication.translate('IPFSSearchResults',
        '<b>No results found</b>')

def iErrFetching():
    return QCoreApplication.translate('IPFSSearchResults',
        '<b>Error while fetching results</b>')

def iSearching():
    return QCoreApplication.translate('IPFSSearchResults',
        '<b>Searching ...</b>')

class IPFSSearchResultsW(QWidget):
    def __init__(self, results, template, parent=None):
        super().__init__(parent)

        self.ui = ui_ipfssearchwresults.Ui_IPFSSearchResults()
        self.ui.setupUi(self)

        self.results = results
        self.searchW = parent
        self.template = template

    def render(self):
        ctxHits = []

        for hit in self.results.hits:
            hitHash = hit.get('hash', None)

            if hitHash is None or not cidValid(hitHash):
                continue

            ctxHits.append({
                'path':  joinIpfs(hitHash),
                'title': hit.get('title', iUnknown()),
                'size':  hit.get('size', iUnknown()),
                'descr': hit.get('description', ''),
                'type':  hit.get('type', iUnknown()),
                'first-seen': hit.get('first-seen', iUnknown())
            })

        pageHtml = self.template.render(hits=ctxHits)

        self.ui.browser.setHtml(pageHtml)
        self.ui.browser.setOpenLinks(False)
        self.ui.browser.setOpenExternalLinks(False)
        self.ui.browser.anchorClicked.connect(self.onAnchorClicked)

        self.highlighter = Highlighter(self.searchW.searchQuery,
                self.ui.browser.document())

    def onAnchorClicked(self, url):
        fragment = url.fragment()
        path = url.path()

        if not path:
            return

        hashV = stripIpfs(path)

        if fragment == 'hashmark':
            hit = self.results.findByHash(hashV)
            title = hit.get('title', iUnknown()) if hit else ''
            descr = hit.get('description', iUnknown()) if hit else ''

            addHashmark(self.searchW.app.marksLocal,
                    path, title, description=descr)

        elif fragment == 'explore':
            mainW = self.searchW.app.mainWindow
            mainW.exploreHash(hashV)
        else:
            tab = self.searchW.gWindow.addBrowserTab()
            tab.enterUrl(url)

class EmptyResultsW(IPFSSearchResultsW):
    def render(self):
        self.ui.browser.setHtml(iNoResults())

class ErrorPage(IPFSSearchResultsW):
    def render(self):
        self.ui.browser.setHtml(iErrFetching())

class Highlighter(QSyntaxHighlighter):
    """
    Highlights the search query string in the results
    """
    def __init__(self, highStr, doc):
        super().__init__(doc)

        self.highStr =  highStr
        self.fmt = QTextCharFormat()
        self.fmt.setFontWeight(QFont.Bold);
        self.fmt.setForeground(Qt.red);

    def highlightBlock(self, text):
        rExpr = QRegularExpression(self.highStr)
        matches = rExpr.globalMatch(text)

        if not matches.isValid():
            return

        while matches.hasNext():
            match = matches.next()
            if not match.isValid():
                break
            self.setFormat(match.capturedStart(), match.capturedLength(),
                    self.fmt)

class IPFSSearchView(GalacteekTab):
    resultsReceived = pyqtSignal(ipfssearch.IPFSSearchResults, bool)

    def __init__(self, searchQuery, *args, **kw):
        super(IPFSSearchView, self).__init__(*args, **kw)

        self.searchQuery = searchQuery
        self.pages = {}
        self.pageCount = 0

        self.ui = ui_ipfssearchw.Ui_IPFSSearchMain()
        self.ui.setupUi(self)

        self.stack = QStackedWidget()
        self.ui.verticalLayout.addWidget(self.stack)
        self.ui.comboPages.activated.connect(self.onComboPages)
        self.ui.prevPageButton.clicked.connect(self.onPrevPage)
        self.ui.nextPageButton.clicked.connect(self.onNextPage)
        self.ui.prevPageButton.setIcon(
                self.style().standardIcon(QStyle.SP_MediaSkipBackward))
        self.ui.nextPageButton.setIcon(
                self.style().standardIcon(QStyle.SP_MediaSkipForward))
        self.stack.currentChanged.connect(self.onPageChanged)
        self.resultsReceived.connect(self.onResultsRx)

        self.resultsTempl = self.app.getJinjaTemplate('ipfssearch.html')

        if self.resultsTempl and self.searchQuery:
            self.app.task(self.runSearch, self.searchQuery)
        else:
            self.addErrorPage()

    @property
    def currentPage(self):
        return self.ui.comboPages.currentIndex()

    def getPageData(self, page):
        return self.pages.get(page, None)

    def getPageWidget(self, page):
        pageData = self.pages.get(page, None)
        if pageData:
            return pageData['stackw']

    def disableCombo(self):
        self.ui.comboPages.setEnabled(False)

    def enableCombo(self):
        self.ui.comboPages.setEnabled(True)

    def enableNavArrows(self, enable=True):
        self.ui.prevPageButton.setEnabled(enable)
        self.ui.nextPageButton.setEnabled(enable)

    def onPageChanged(self, idx):
        self.ui.prevPageButton.setEnabled(idx > 0)
        self.ui.nextPageButton.setEnabled(self.pageCount > idx+1)

    def onPrevPage(self):
        cIdx = self.currentPage
        self.loadPage(cIdx-1)

    def onNextPage(self):
        cIdx = self.currentPage
        self.loadPage(cIdx+1)

    def onComboPages(self, idx):
        self.loadPage(idx)

    def loadPage(self, pageNum):
        self.disableCombo()

        pageData = self.getPageData(pageNum)

        if not pageData:
            self.app.task(self.runSearchPage, self.searchQuery,
                pageNum, True)
        else:
            self.displayPage(pageNum)
            self.enableCombo()

    def addEmptyResultsPage(self, searchR):
        page = EmptyResultsW(searchR, self.resultsTempl, parent=self)
        page.render()
        return self.stack.addWidget(page), page

    def addErrorPage(self):
        page = EmptyResultsW(None, self.resultsTempl, parent=self)
        page.render()
        return self.stack.addWidget(page), page

    def addResultsPage(self, searchR):
        page = IPFSSearchResultsW(searchR, self.resultsTempl, parent=self)
        page.render()
        return self.stack.addWidget(page), page

    async def runSearch(self, searchQuery):
        self.ui.labelInfo.setText(iSearching())
        self.ui.comboPages.clear()

        async for sr in ipfssearch.search(searchQuery, preloadPages=3):
            await asyncio.sleep(0)
            pageCount = sr.pageCount

            if pageCount == 0:
                self.addEmptyResultsPage(sr)
                self.enableNavArrows(False)
                self.disableCombo()
                break

            if sr.page == 0 and self.ui.comboPages.count() == 0:
                resCount = sr.results.get('total', iUnknown())
                maxScore = sr.results.get('max_score', iUnknown())
                self.ui.labelInfo.setText(iResultsInfo(resCount, maxScore))

                if pageCount > 256:
                    pageCount = 256

                self.pageCount = pageCount

                for pageNum in range(1, pageCount+1):
                    self.ui.comboPages.insertItem(pageNum,
                        'Page {}'.format(pageNum))

            self.resultsReceived.emit(sr, False)

    async def runSearchPage(self, searchQuery, page, display=False):
        sr = await ipfssearch.getPageResults(searchQuery, page)
        if sr:
            self.resultsReceived.emit(sr, display)

        self.enableCombo()

    def displayPage(self, page):
        pageW = self.getPageWidget(page)
        if pageW:
            self.stack.setCurrentWidget(pageW)
            self.ui.comboPages.setCurrentIndex(page)

    def onResultsRx(self, sr, display):
        idx, page = self.addResultsPage(sr)
        self.pages[sr.page] = {
                'results': sr,
                'stackidx': idx,
                'stackw': page
        }
        if display:
            self.displayPage(sr.page)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/addkeydialog.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AddKeyDialog(object):
    def setupUi(self, AddKeyDialog):
        AddKeyDialog.setObjectName("AddKeyDialog")
        AddKeyDialog.resize(502, 145)
        self.gridLayout_2 = QtWidgets.QGridLayout(AddKeyDialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.label = QtWidgets.QLabel(AddKeyDialog)
        self.label.setObjectName("label")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.label_2 = QtWidgets.QLabel(AddKeyDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.keyName = QtWidgets.QLineEdit(AddKeyDialog)
        self.keyName.setMaxLength(64)
        self.keyName.setObjectName("keyName")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.keyName)
        self.keySize = QtWidgets.QComboBox(AddKeyDialog)
        self.keySize.setObjectName("keySize")
        self.keySize.addItem("")
        self.keySize.addItem("")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.keySize)
        self.gridLayout_2.addLayout(self.formLayout_2, 0, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(AddKeyDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout_2.addWidget(self.buttonBox, 1, 0, 1, 1)

        self.retranslateUi(AddKeyDialog)
        self.buttonBox.accepted.connect(AddKeyDialog.accept)
        self.buttonBox.rejected.connect(AddKeyDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(AddKeyDialog)

    def retranslateUi(self, AddKeyDialog):
        _translate = QtCore.QCoreApplication.translate
        AddKeyDialog.setWindowTitle(_translate("AddKeyDialog", "Dialog"))
        self.label.setText(_translate("AddKeyDialog", "Key name"))
        self.label_2.setText(_translate("AddKeyDialog", "Key size"))
        self.keySize.setItemText(0, _translate("AddKeyDialog", "2048"))
        self.keySize.setItemText(1, _translate("AddKeyDialog", "4096"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/mediaplaylist.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MediaPlaylist(object):
    def setupUi(self, MediaPlaylist):
        MediaPlaylist.setObjectName("MediaPlaylist")
        MediaPlaylist.resize(441, 328)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(MediaPlaylist)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.savePlaylistButton = QtWidgets.QToolButton(MediaPlaylist)
        self.savePlaylistButton.setObjectName("savePlaylistButton")
        self.horizontalLayout.addWidget(self.savePlaylistButton)
        self.loadPlaylistButton = QtWidgets.QToolButton(MediaPlaylist)
        self.loadPlaylistButton.setObjectName("loadPlaylistButton")
        self.horizontalLayout.addWidget(self.loadPlaylistButton)
        self.clearButton = QtWidgets.QToolButton(MediaPlaylist)
        self.clearButton.setObjectName("clearButton")
        self.horizontalLayout.addWidget(self.clearButton)
        self.clipPlaylistButton = QtWidgets.QToolButton(MediaPlaylist)
        self.clipPlaylistButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/clipboard.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.clipPlaylistButton.setIcon(icon)
        self.clipPlaylistButton.setObjectName("clipPlaylistButton")
        self.horizontalLayout.addWidget(self.clipPlaylistButton)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.previousButton = QtWidgets.QPushButton(MediaPlaylist)
        self.previousButton.setText("")
        self.previousButton.setObjectName("previousButton")
        self.horizontalLayout_2.addWidget(self.previousButton)
        self.nextButton = QtWidgets.QPushButton(MediaPlaylist)
        self.nextButton.setText("")
        self.nextButton.setObjectName("nextButton")
        self.horizontalLayout_2.addWidget(self.nextButton)
        self.horizontalLayout.addLayout(self.horizontalLayout_2)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.listView = QtWidgets.QListView(MediaPlaylist)
        self.listView.setObjectName("listView")
        self.verticalLayout_3.addWidget(self.listView)
        self.verticalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(MediaPlaylist)
        QtCore.QMetaObject.connectSlotsByName(MediaPlaylist)

    def retranslateUi(self, MediaPlaylist):
        _translate = QtCore.QCoreApplication.translate
        MediaPlaylist.setWindowTitle(_translate("MediaPlaylist", "Form"))
        self.savePlaylistButton.setText(_translate("MediaPlaylist", "Save playlist"))
        self.loadPlaylistButton.setText(_translate("MediaPlaylist", "Load playlist"))
        self.clearButton.setText(_translate("MediaPlaylist", "Clear"))
        self.previousButton.setToolTip(_translate("MediaPlaylist", "<html><head/><body><p>Previous</p></body></html>"))
        self.nextButton.setToolTip(_translate("MediaPlaylist", "<html><head/><body><p>Next</p></body></html>"))

from . import galacteek_rc




from PyQt5.QtCore import QCoreApplication, QUrl, QStandardPaths, QSettings

# Settings sections
CFG_SECTION_IPFSD = 'ipfsdaemon'
CFG_SECTION_IPFS = 'ipfs'
CFG_SECTION_BROWSER = 'browser'
CFG_SECTION_IPFSCONN1 = 'ipfsconn1'
CFG_SECTION_UI = 'ui'
CFG_SECTION_USERINFO = 'userinfo'

CFG_KEY_ENABLED = 'enabled'

# These keys are for ipfsdaemon and ipfsconn sections
CFG_KEY_APIPORT = 'apiport'
CFG_KEY_SWARMPORT = 'swarmport'
CFG_KEY_HTTPGWPORT = 'httpgwport'
CFG_KEY_HOST = 'host'
CFG_KEY_SWARMLOWWATER = 'swarm_lowwater'
CFG_KEY_SWARMHIGHWATER = 'swarm_highwater'
CFG_KEY_STORAGEMAX = 'storagemax' # integer, max storage in Gb
CFG_KEY_CORS = 'cors'

# Browser
CFG_KEY_HOMEURL = 'homeurl'
CFG_KEY_GOTOHOME = 'gotohomeonnewtab'
CFG_KEY_DLPATH = 'downloadspath'
CFG_KEY_ALLOWHTTPBROWSING = 'httpbrowsing'
CFG_KEY_JSAPI = 'jsapi'
CFG_KEY_PPAPIPLUGINS = 'ppapiplugins'

# IPFS
CFG_KEY_PUBSUB = 'pubsub'
CFG_KEY_HASHMARKSEXCH = 'hashmarksexch'

# UI
CFG_KEY_WRAPSINGLEFILES = 'wrapsinglefiles'
CFG_KEY_WRAPDIRECTORIES = 'wrapdirectories'
CFG_KEY_HIDEHASHES = 'hidehashes'
CFG_KEY_LANG = 'lang'

# for fast access
S_HOMEURL = (CFG_SECTION_BROWSER, CFG_KEY_HOMEURL)
S_GOTOHOME = (CFG_SECTION_BROWSER, CFG_KEY_GOTOHOME)
S_DOWNLOADS_PATH = (CFG_SECTION_BROWSER, CFG_KEY_DLPATH)

def setDefaultSettings(gApp):
    # Sets the default settings
    sManager = gApp.settingsMgr

    section = CFG_SECTION_IPFSD
    sManager.setDefaultSetting(section, CFG_KEY_APIPORT, 5001)
    sManager.setDefaultSetting(section, CFG_KEY_SWARMPORT, 4001)
    sManager.setDefaultSetting(section, CFG_KEY_HTTPGWPORT, 8080)
    sManager.setDefaultSetting(section, CFG_KEY_SWARMHIGHWATER, 300)
    sManager.setDefaultSetting(section, CFG_KEY_SWARMLOWWATER, 150)
    sManager.setDefaultSetting(section, CFG_KEY_STORAGEMAX, 50)
    sManager.setDefaultTrue(section, CFG_KEY_CORS)
    sManager.setDefaultTrue(section, CFG_KEY_ENABLED)

    section = CFG_SECTION_BROWSER
    sManager.setDefaultSetting(section, CFG_KEY_HOMEURL, 'ipfs:/ipns/ipfs.io')
    sManager.setDefaultSetting(section, CFG_KEY_DLPATH,
        gApp.defaultDownloadsLocation)
    sManager.setDefaultTrue(section, CFG_KEY_GOTOHOME)
    sManager.setDefaultFalse(section, CFG_KEY_JSAPI)
    sManager.setDefaultFalse(section, CFG_KEY_ALLOWHTTPBROWSING)
    sManager.setDefaultTrue(section, CFG_KEY_PPAPIPLUGINS)

    # Default IPFS connection when not spawning daemon
    section = CFG_SECTION_IPFSCONN1
    sManager.setDefaultSetting(section, CFG_KEY_HOST, 'localhost')
    sManager.setDefaultSetting(section, CFG_KEY_APIPORT, 5001)
    sManager.setDefaultSetting(section, CFG_KEY_HTTPGWPORT, 8080)

    section = CFG_SECTION_IPFS
    sManager.setDefaultTrue(section, CFG_KEY_PUBSUB)
    sManager.setDefaultTrue(section, CFG_KEY_HASHMARKSEXCH)

    section = CFG_SECTION_UI
    sManager.setDefaultTrue(section, CFG_KEY_WRAPSINGLEFILES)
    sManager.setDefaultFalse(section, CFG_KEY_WRAPDIRECTORIES)
    sManager.setDefaultFalse(section, CFG_KEY_HIDEHASHES)
    sManager.setDefaultSetting(section, CFG_KEY_LANG, 'en')

    sManager.sync()
    return True

class SettingsManager(object):
    # QSettings has its problems with pyqt5 regarding booleans
    trueVal = 'true'
    falseVal = 'false'

    def __init__(self, path=None):
        self.settings = QSettings(path, QSettings.IniFormat)

    def sync(self):
        """ Synchronize settings to disk """
        return self.settings.sync()

    def eSet(self, key, value):
        """ Easy set. key is a (section, key) tuple """
        return self.setSetting(key[0], key[1], value)

    def eGet(self, key, type=None):
        """ Easy get. key is a (section, key) tuple """
        return self.getSetting(key[0], key[1], type=type)

    def setSetting(self, section, name, value):
        """ Changes a setting to given value

        :param str section: setting section
        :param str name: key inside the section
        :param value: setting value
        """
        self.settings.setValue('{0}/{1}'.format(section, name), value)

    def setDefaultSetting(self, section, name, value):
        """
        Sets default setting for section and name if that setting hasn't
        been registered yet
        """
        existing = self.getSetting(section, name)
        if not existing:
            self.settings.setValue('{0}/{1}'.format(section, name), value)

    def setDefaultTrue(self, section, name):
        return self.setDefaultSetting(section, name, self.trueVal)

    def setDefaultFalse(self, section, name):
        return self.setDefaultSetting(section, name, self.falseVal)

    def getInt(self, section, name):
        """
        Gets the setting referenced by section and name, casting it as
        an integer
        """
        return self.getSetting(section, name, type=int)

    def getSetting(self, section, name, type=None):
        """ Gets a setting, casting it to the given type

        :param str section: setting section
        :param str name: key inside the section
        :param type: type to convert to
        """
        key = '{0}/{1}'.format(section, name)
        if type:
            return self.settings.value(key, type=type)
        else:
            return self.settings.value(key)

    def setTrue(self, section, name):
        return self.setSetting(section, name, self.trueVal)

    def setFalse(self, section, name):
        return self.setSetting(section, name, self.falseVal)

    def setBoolFrom(self, section, name, boolVal):
        if boolVal is True:
            self.setTrue(section, name)
        elif boolVal is False:
            self.setFalse(section, name)

    def isTrue(self, section, name):
        return self.getSetting(section, name) == self.trueVal

    def isFalse(self, section, name):
        return self.getSetting(section, name) == self.falseVal

    # Properties

    @property
    def hideHashes(self):
        return self.isTrue(CFG_SECTION_UI, CFG_KEY_HIDEHASHES)

    @property
    def wrapFiles(self):
        return self.isTrue(CFG_SECTION_UI, CFG_KEY_WRAPSINGLEFILES)

    @property
    def wrapDirectories(self):
        return self.isTrue(CFG_SECTION_UI, CFG_KEY_WRAPDIRECTORIES)

    @property
    def allowHttpBrowsing(self):
        return self.isTrue(CFG_SECTION_BROWSER, CFG_KEY_ALLOWHTTPBROWSING)

    @property
    def jsIpfsApi(self):
        return self.isTrue(CFG_SECTION_BROWSER, CFG_KEY_JSAPI)

    @property
    def ppApiPlugins(self):
        return self.isTrue(CFG_SECTION_BROWSER, CFG_KEY_PPAPIPLUGINS)




import sys
import time
import os.path
import mimetypes

from PyQt5.QtWidgets import (QWidget, QFrame, QApplication,
        QLabel, QPushButton, QVBoxLayout, QAction, QHBoxLayout,
        QTreeView, QHeaderView, QShortcut, QToolBox, QTextBrowser,
        QFileDialog, QProgressBar, QSpacerItem, QSizePolicy,
        QToolButton, QAbstractItemView)

from PyQt5.QtGui import QStandardItemModel, QStandardItem, QKeySequence, QIcon

from PyQt5.QtCore import (QCoreApplication, QUrl, Qt, QEvent, QObject,
    pyqtSignal, QMimeData, QMimeDatabase)

from galacteek.appsettings import *
from galacteek.ipfs import cidhelpers
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp, ipfsStatOp
from galacteek.ipfs.cache import IPFSEntryCache

from . import galacteek_rc
from . import modelhelpers
from .i18n import *
from .helpers import *
from .widgets import GalacteekTab
from .hashmarks import *

import aioipfs

def iCIDInfo(cidv, linksCount, size):
    return QCoreApplication.translate('IPFSHashExplorer',
        'CID (v{0}): {1} links, total size: {2}').format(
            cidv, linksCount, size)

def iGitClone():
    return QCoreApplication.translate('IPFSHashExplorer',
        'Clone repository')

def iGitClonedRepo(path):
    return QCoreApplication.translate('IPFSHashExplorer',
        'Cloned git repository: {0}').format(path)

def iGitErrorCloning(msg):
    return QCoreApplication.translate('IPFSHashExplorer',
        'Error cloning git repository: {0}').format(msg)

def iGitInvalid():
    return QCoreApplication.translate('IPFSHashExplorer',
        'Invalid git repository')

def iLoading():
    return QCoreApplication.translate('IPFSHashExplorer', 'Loading ...')

def iTimeout():
    return QCoreApplication.translate('IPFSHashExplorer',
            'Timeout error')

def iTimeoutTryNoResolve():
    return QCoreApplication.translate('IPFSHashExplorer',
            'Timeout error: trying without resolving nodes types ..')

def iTimeoutInvalidHash():
    return QCoreApplication.translate('IPFSHashExplorer',
            'Timeout error: invalid hash')

class IPFSItem(UneditableItem):
    def __init__(self, text, icon=None):
        super().__init__(text, icon=icon)
        self.setParentHash(None)

    def setParentHash(self, hash):
        self.parentHash = hash

    def getParentHash(self):
        return self.parentHash

class IPFSNameItem(IPFSItem):
    def __init__(self, entry, text, icon):
        super().__init__(text, icon=icon)

        self._entry = entry
        self._mimeType = None

    @property
    def entry(self):
        return self._entry

    @property
    def mimeType(self):
        return self._mimeType

    def mimeFromDb(self, db):
        self._mimeType = db.mimeTypeForFile(self.entry['Name'])

    @property
    def mimeCategory(self):
        if self.mimeType:
            return self.mimeTypeName.split('/')[0]

    @property
    def mimeTypeName(self):
        if self.mimeType:
            return self.mimeType.name()

    def cid(self):
        return cidhelpers.getCID(self.entry['Hash'])

    def getFullPath(self):
        """
        Returns the full IPFS path of the entry associated with this item
        (preserving file names) if we have the parent's hash, or the IPFS path
        with the entry's hash otherwise
        """
        parentHash = self.getParentHash()
        name = self.entry['Name']
        if parentHash:
            return joinIpfs(os.path.join(parentHash, name))
        else:
            return joinIpfs(self.entry['Hash'])

    def isRaw(self):
        return self.entry['Type'] == 0

    def isDir(self):
        return self.entry['Type'] == 1

    def isFile(self):
        return self.entry['Type'] == 2

    def isMetadata(self):
        return self.entry['Type'] == 3

    def isSymlink(self):
        return self.entry['Type'] == 4

    def isUnknown(self):
        return self.entry['Type'] == -1

class IPFSHashItemModel(QStandardItemModel):
    COL_NAME = 0
    COL_SIZE = 1
    COL_MIME = 2
    COL_HASH = 3

    def __init__(self, parent, *args, **kw):
        QStandardItemModel.__init__(self, *args, **kw)

        self.entryCache = IPFSEntryCache()
        self.rowsInserted.connect(self.onRowsInserted)

    def canDropMimeData(self, data, action, row, column, parent):
        return False

    def getHashFromIdx(self, idx):
        idxHash = self.index(idx.row(), self.COL_HASH, idx.parent())
        return self.data(idxHash)

    def getNameFromIdx(self, idx):
        idxName = self.index(idx.row(), self.COL_NAME, idx.parent())
        return self.data(idxName)

    def getNameItemFromIdx(self, idx):
        idxName = self.index(idx.row(), self.COL_NAME, idx.parent())
        return self.itemFromIndex(idxName)

    def onRowsInserted(self, parent, first, last):
        """ Update the entry cache when rows are added """
        for itNum in range(first, last+1):
            itNameIdx = self.index(itNum, self.COL_NAME, parent)
            itName = self.itemFromIndex(itNameIdx)
            entry = itName.entry
            self.entryCache.register(entry)

class IPFSHashExplorerToolBox(GalacteekTab):
    """
    Organizes IPFSHashExplorerWidgets with a QToolBox
    """
    def __init__(self, gWindow, hashRef, maxItems=16, parent=None):
        super(IPFSHashExplorerToolBox, self).__init__(gWindow)

        self.rootHash = hashRef
        self.maxItems = maxItems

        self.toolbox = QToolBox()
        self.vLayout = QVBoxLayout(self)
        self.vLayout.addWidget(self.toolbox)

        if self.rootHash:
            self.viewHash(self.rootHash)

    @property
    def itemsCount(self):
        return self.toolbox.count()

    def viewHash(self, hashRef, addClose=False, autoOpenFolders=False):
        w = self.lookup(hashRef)
        if w:
            self.toolbox.setCurrentWidget(w)
            return True

        if self.itemsCount > self.maxItems:
            return False

        view = IPFSHashExplorerWidget(self.gWindow, hashRef,
                parent=self, addClose=addClose,
                autoOpenFolders=autoOpenFolders)
        idx = self.toolbox.addItem(view, getIconIpfsWhite(), hashRef)
        self.toolbox.setCurrentIndex(idx)
        view.reFocus()
        return True

    def lookup(self, hashRef):
        for idx in range(self.itemsCount):
            if self.toolbox.itemText(idx) == hashRef:
                return self.toolbox.widget(idx)

    def remove(self, view):
        idx = self.toolbox.indexOf(view)
        if idx:
            self.toolbox.removeItem(idx)
            # Always display previous index in the stack
            if self.itemsCount > 0:
                rIdx = idx - 1
                view = self.toolbox.widget(rIdx)
                if view:
                    self.toolbox.setCurrentWidget(view)
                    view.reFocus()

class TreeEventFilter(QObject):
    copyPressed = pyqtSignal()
    returnPressed = pyqtSignal()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if key == Qt.Key_Return:
                self.returnPressed.emit()
                return True
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_C:
                    self.copyPressed.emit()
                    return True
        return False

class TextView(GalacteekTab):
    def __init__(self, data, mimeType, parent=None):
        super(TextView, self).__init__(parent)

        self.textData = self.decode(data)

        if not self.textData:
            self.textData = 'Error decoding data'

        self.vLayout = QVBoxLayout(self)

        self.textBrowser = QTextBrowser()
        if mimeType == 'text/html':
            self.textBrowser.setHtml(self.textData)
        else:
            self.textBrowser.setPlainText(self.textData)

        self.vLayout.addWidget(self.textBrowser)
        self.setLayout(self.vLayout)

    def decode(self, data):
        for enc in ['utf-8', 'latin1', 'ascii']:
            try:
                textData = data.decode(enc)
            except:
                continue
            else:
                return textData

class HashTreeView(QTreeView):
    pass

class IPFSHashExplorerWidget(QWidget):
    def __init__(self, gWindow, hashRef, addClose=False,
            autoOpenFolders=False, parent=None):
        super(IPFSHashExplorerWidget, self).__init__(parent)

        self.parent = parent

        self.gWindow = gWindow
        self.app = gWindow.app
        self.rootHash = hashRef
        self.rootPath = joinIpfs(self.rootHash)
        self.cid = cidhelpers.getCID(self.rootHash)

        self.vLayout = QVBoxLayout(self)

        self.autoOpenFolders = autoOpenFolders
        self.hLayoutTop = QHBoxLayout()
        self.hLayoutInfo = QHBoxLayout()
        self.hLayoutCtrl = QHBoxLayout()
        self.hLayoutTop.addLayout(self.hLayoutInfo)
        spacerItem = QSpacerItem(10, 20, QSizePolicy.Expanding,
                QSizePolicy.Minimum)
        self.hLayoutTop.addItem(spacerItem)
        self.hLayoutTop.addLayout(self.hLayoutCtrl)

        self.labelInfo = QLabel()

        self.hLayoutInfo.addWidget(self.labelInfo, 0, Qt.AlignLeft)

        if addClose:
            self.closeButton = QPushButton('Close')
            self.closeButton.clicked.connect(self.onCloseView)
            self.closeButton.setMaximumWidth(100)
            self.closeButton.setShortcut(QKeySequence('Ctrl+w'))
            self.hLayoutCtrl.addWidget(self.closeButton, 0, Qt.AlignLeft)

        self.getTask = None
        self.getButton = QPushButton(iDownload())
        self.getButton.clicked.connect(self.onGet)
        self.getButton.setShortcut(QKeySequence('Ctrl+d'))
        self.getLabel = QLabel()
        self.getProgress = QProgressBar()
        self.getProgress.setMinimum(0)
        self.getProgress.setMaximum(100)
        self.getProgress.hide()

        self.markButton = QPushButton(getIcon('hashmarks.png'), iHashmark())
        self.markButton.clicked.connect(self.onHashmark)

        self.pinButton = QPushButton('Pin')
        self.pinButton.clicked.connect(self.onPin)
        self.pinButton.setShortcut(QKeySequence('Ctrl+p'))

        self.hLayoutCtrl.addWidget(self.getButton)
        self.hLayoutCtrl.addWidget(self.pinButton)
        self.hLayoutCtrl.addWidget(self.markButton)
        self.hLayoutCtrl.addWidget(self.getLabel)
        self.hLayoutCtrl.addWidget(self.getProgress)

        self.vLayout.addLayout(self.hLayoutTop)

        self.tree = HashTreeView()
        self.tree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.tree.setDragDropMode(QAbstractItemView.DragOnly)
        self.vLayout.addWidget(self.tree)

        self.model = IPFSHashItemModel(self)
        self.model.setHorizontalHeaderLabels(
                [iFileName(), iFileSize(), iMimeType(), iFileHash()])
        self.itemRoot = self.model.invisibleRootItem()
        self.itemRootIdx = self.model.indexFromItem(self.itemRoot)

        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.onContextMenu)
        self.tree.doubleClicked.connect(self.onDoubleClicked)
        self.tree.setSortingEnabled(True)

        evfilter = IPFSTreeKeyFilter(self.tree)
        evfilter.copyHashPressed.connect(self.onCopyItemHash)
        evfilter.copyPathPressed.connect(self.onCopyItemPath)
        evfilter.returnPressed.connect(self.onReturnPressed)
        self.tree.installEventFilter(evfilter)
        self.iconFolder = getIcon('folder-open.png')
        self.iconFile = getIcon('file.png')
        self.iconUnknown = getIcon('unknown-file.png')

        self.updateTree()

    def setInfo(self, text):
        self.labelInfo.setText(text)

    def reFocus(self):
        self.tree.setFocus(Qt.OtherFocusReason)

    @ipfsOp
    async def gitClone(self, ipfsop, entry, dest):
        """
        Clone the git repository contained within entry, to directory dest
        """
        from git.repo import base
        from git.exc import InvalidGitRepositoryError

        self.gitButton.setEnabled(False)

        getRet = await ipfsop.client.get(entry['Hash'],
                dstdir=self.app.tempDir.path())
        if getRet != True:
            self.gitButton.setEnabled(True)
            return messageBox('Could not fetch the git repository')

        repoPath = os.path.join(self.app.tempDir.path(), entry['Hash'])
        try:
            repo = base.Repo(repoPath)
        except InvalidGitRepositoryError:
            return messageBox(iGitInvalid())

        dstPath = '{}.git'.format(
            os.path.join(dest, self.rootHash))
        # Clone it now. No need to run it in a threadpool since the git module
        # will run a git subprocess for the cloning
        try:
            newRepo = repo.clone(dstPath)
        except Exception as e:
            self.gitButton.setEnabled(True)
            return messageBox(iGitErrorCloning(str(e)))

        messageBox(iGitClonedRepo(dstPath))
        self.gitButton.setEnabled(True)

    def addGitControl(self, gitEntry):
        """
        Adds a tool button for making operations on the git repo
        """
        def clone(entry):
            dirSel = directorySelect()
            if dirSel:
                self.app.task(self.gitClone, entry, dirSel)

        self.gitMenu = QMenu()
        self.gitMenu.addAction(iGitClone(), lambda: clone(gitEntry))
        self.gitButton = QToolButton()
        self.gitButton.setText('Git')
        self.gitButton.setMenu(self.gitMenu)
        self.gitButton.setPopupMode(QToolButton.MenuButtonPopup)
        self.hLayoutCtrl.addWidget(self.gitButton)

    def onHashmark(self):
        addHashmark(self.app.marksLocal,
            self.rootPath, '',
            stats=self.app.ipfsCtx.objectStats.get(
                self.rootPath, {}))

    def onGet(self):
        dirSel = directorySelect()
        if dirSel:
            self.getTask = self.app.task(self.getResource, self.rootPath,
                    dirSel)

    def onPin(self):
        self.app.task(self.app.ipfsCtx.pinner.queue, self.rootPath, True,
                None)

    def onReturnPressed(self):
        currentIdx = self.tree.currentIndex()
        if currentIdx.isValid():
            self.onDoubleClicked(currentIdx)

    def onCopyItemHash(self):
        dataHash = self.model.getHashFromIdx(self.tree.currentIndex())
        self.app.setClipboardText(dataHash)

    def onCopyItemPath(self):
        dataHash = self.model.getHashFromIdx(self.tree.currentIndex())
        self.app.setClipboardText(joinIpfs(dataHash))

    def onCloseView(self):
        self.parent.remove(self)

    def browse(self, hash):
        self.gWindow.addBrowserTab().browseIpfsHash(hash)

    def browseFs(self, path):
        self.gWindow.addBrowserTab().browseFsPath(path)

    def onDoubleClicked(self, idx):
        nameItem = self.model.getNameItemFromIdx(idx)
        dataHash = self.model.getHashFromIdx(idx)

        if nameItem.isDir():
            self.parent.viewHash(dataHash, addClose=True)
        elif nameItem.isFile() or nameItem.isUnknown():
            self.openFile(nameItem, dataHash)

    def openFile(self, item, fileHash):
        if item.mimeTypeName is None or item.mimeTypeName == 'text/html':
            return self.browse(fileHash)

        self.gWindow.app.task(self.openFileWithMime, item, fileHash)

    @ipfsOp
    async def openFileWithMime(self, ipfsop, item, fileHash):
        fullPath = item.getFullPath()

        if item.mimeCategory == 'text':
            data = await ipfsop.client.cat(fileHash)
            tView = TextView(data, item.mimeTypeName, parent=self.gWindow)
            self.gWindow.registerTab(tView, item.entry['Name'], current=True)
        elif item.mimeCategory == 'video' or item.mimeCategory == 'audio':
            return self.gWindow.mediaPlayerQueue(fullPath)
        elif item.mimeCategory == 'image':
            return self.browseFs(fullPath)
        else:
            # Default
            return self.browseFs(fullPath)

    def updateTree(self):
        self.app.task(self.listHash, self.rootPath, parentItem=self.itemRoot)

    def onContextMenu(self, point):
        selModel = self.tree.selectionModel()
        rows = selModel.selectedRows()

        items = [self.model.getNameItemFromIdx(idx) for idx in rows]

        menu = QMenu()

        def pinRecursive():
            for item in items:
                fp = item.getFullPath()
                self.app.task(self.app.ipfsCtx.pinner.queue, fp, True, None)

        def queueMedia():
            for item in items:
                self.gWindow.mediaPlayerQueue(item.getFullPath())

        def download():
            dirSel = directorySelect()
            for item in items:
                getTask = self.app.task(self.getResource, item.getFullPath(),
                    dirSel)

        menu.addAction(getIcon('pin-black.png'), 'Pin (recursive)',
                lambda: pinRecursive())
        menu.addAction(getIcon('multimedia.png'), 'Queue in media player',
                lambda: queueMedia())
        menu.addAction('Download', lambda: download())

        menu.exec(self.tree.mapToGlobal(point))

    async def timedList(self, ipfsop, objPath, parentItem, autoexpand,
            secs, resolve_type):
        return await asyncio.wait_for(
            self.list(ipfsop, objPath,
                parentItem=parentItem, autoexpand=autoexpand,
                resolve_type=resolve_type), secs)

    @ipfsOp
    async def listHash(self, ipfsop, objPath, parentItem,
            autoexpand=False):
        """ Lists contents of IPFS object referenced by objPath,
            and change the tree's model afterwards.

            We first try with resolve-type set to true, and if that gets a
            timeout do the same call but without resolving nodes types
        """

        try:
            self.setInfo(iLoading())
            await self.timedList(ipfsop, objPath, parentItem,
                    autoexpand, 15, True)
        except asyncio.TimeoutError as timeoutE:
            self.setInfo(iTimeoutTryNoResolve())

            try:
                await self.timedList(ipfsop, objPath, parentItem,
                    autoexpand, 10, False)
            except asyncio.TimeoutError as timeoutE:
                # That's a dead end .. bury that hash please ..
                self.setInfo(iTimeoutInvalidHash())
                return

        except aioipfs.APIError:
            self.setInfo(iErrNoCx())
            return

        self.tree.setModel(self.model)
        self.tree.header().setSectionResizeMode(self.model.COL_NAME,
                QHeaderView.ResizeToContents)
        self.tree.header().setSectionResizeMode(self.model.COL_MIME,
                QHeaderView.ResizeToContents)
        self.tree.sortByColumn(self.model.COL_NAME, Qt.AscendingOrder)

        if self.app.settingsMgr.hideHashes:
            self.tree.hideColumn(self.model.COL_HASH)

        rStat = await ipfsop.objStatCtxUpdate(objPath)

        if rStat:
            self.setInfo(iCIDInfo(self.cid.version,
                rStat['NumLinks'],
                sizeFormat(rStat['CumulativeSize'])))

    async def list(self, op, path, parentItem=None,
            autoexpand=False, resolve_type=True):

        parentItemSibling = self.model.sibling(parentItem.row(),
                self.model.COL_HASH,
                parentItem.index())
        parentItemHash = self.model.data(parentItemSibling)
        if parentItemHash is None:
            parentItemHash = self.rootHash

        async for obj in op.list(path, resolve_type):
            for entry in obj['Links']:
                await op.sleep()

                if entry['Hash'] in self.model.entryCache:
                    continue

                nItemName = IPFSNameItem(entry, entry['Name'], None)
                nItemName.mimeFromDb(self.app.mimeDb)
                nItemName.setParentHash(parentItemHash)
                nItemSize = IPFSItem(sizeFormat(entry['Size']))
                nItemSize.setToolTip(str(entry['Size']))
                nItemMime = IPFSItem(nItemName.mimeTypeName or iUnknown())
                nItemHash = IPFSItem(entry['Hash'])
                nItemName.setToolTip(entry['Hash'])

                if nItemName.isDir():
                    nItemName.setIcon(self.iconFolder)
                elif nItemName.isFile():
                    nItemName.setIcon(self.iconFile)
                elif nItemName.isUnknown():
                    nItemName.setIcon(self.iconUnknown)

                nItem = [nItemName, nItemSize, nItemMime, nItemHash]
                parentItem.appendRow(nItem)

                if nItemName.isDir() and self.autoOpenFolders:
                    # Automatically open sub folders. Used by unit tests
                    self.parent.viewHash(entry['Hash'],
                        addClose=True, autoOpenFolders=self.autoOpenFolders)

                if nItemName.isDir() and entry['Name'] == '.git':
                    # If there's a git repo here, add a control button
                    self.addGitControl(entry)

    @ipfsStatOp
    async def getResource(self, ipfsop, rPath, dest, rStat):
        """
        Get the resource referenced by rPath to directory dest
        """

        if rStat is None:
            return

        self.getProgress.show()
        cumulative = rStat['CumulativeSize']

        async def onGetProgress(ref, bytesRead, arg):
            per = int((bytesRead*100) / cumulative)
            self.getLabel.setText('Downloaded: {0}'.format(
                sizeFormat(bytesRead)))
            self.getProgress.setValue(per)

        ret = await ipfsop.client.get(rPath, dstdir=dest,
                progress_callback=onGetProgress,
                chunk_size=32768)

        self.getLabel.setText('Download finished')
        self.getProgress.hide()



"""
Protobuf definitions for IPFS
"""

from . import unixfs_pb2

from google.protobuf import message
from google.protobuf.json_format import *
from google.protobuf import message_factory

from .protobuf_to_dict import protobuf_to_dict

def decodeDagNode(data):
    try:
        message = unixfs_pb2.Data()
        message.MergeFromString(
                base64.b64decode(data))
        obj = protobuf_to_dict(message)
    except Exception as e:
        return None

    return {
        'type': obj['Type'],
        'size': obj['filesize'],
        'data': obj['Data']
    }




import time
import json
import asyncio

from PyQt5.QtWidgets import (QTextEdit, QVBoxLayout, QHBoxLayout,
        QCheckBox, QPushButton)
from PyQt5.QtCore import QCoreApplication, Qt, QSaveFile, QIODevice

from galacteek import log
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ui.widgets import GalacteekTab
from galacteek.ui.helpers import *
from galacteek.core import jtraverse, asynclib

class EventLogWidget(GalacteekTab):
    """
    Widget to display IPFS log events
    """

    def __init__(self, gWindow):
        super().__init__(gWindow)

        self.logZone = QTextEdit()
        self.logZone.setReadOnly(True)

        self.saveButton = QPushButton('Save')
        self.saveButton.clicked.connect(self.onSave)
        self.clearButton = QPushButton('Clear')
        self.clearButton.clicked.connect(lambda: self.logZone.clear())

        layout = QVBoxLayout(self)
        hLayout = QHBoxLayout(self)

        self.checkCore = QCheckBox('Core events')
        self.checkCore.setCheckState(Qt.Checked)
        self.checkDht = QCheckBox('DHT events')
        self.checkBitswap = QCheckBox('Bitswap events')
        self.checkAll = QCheckBox('All events')

        hLayout.addWidget(self.checkCore)
        hLayout.addWidget(self.checkDht)
        hLayout.addWidget(self.checkBitswap)
        hLayout.addWidget(self.checkAll)
        hLayout.addWidget(self.clearButton)
        hLayout.addWidget(self.saveButton)

        layout.addLayout(hLayout)
        layout.addWidget(self.logZone)

        self.tskLog = self.app.task(self.logWatch)

    def onSave(self):
        fPath = saveFileSelect()
        if fPath:
            file = QSaveFile(fPath)

            if not file.open(QIODevice.WriteOnly):
                return messageBox('Cannot open file for writing')

            file.write(self.logZone.toPlainText().encode())
            file.commit()

    def displayEvent(self, event):
        self.logZone.append(json.dumps(event, indent=4))

    @ipfsOp
    async def logWatch(self, op):
        try:
            await op.client.log.level('all', 'info')

            async for event in op.client.log.tail():
                display = False
                parser = jtraverse.traverseParser(event)

                systems = [ parser.traverse('system'),
                    parser.traverse('Tags.system') ]

                if ('core' in systems or 'addrutil' in systems) and \
                        self.checkCore.isChecked():
                    display = True
                elif 'dht' in systems and self.checkDht.isChecked():
                    display = True
                elif 'bitswap' in systems and self.checkBitswap.isChecked():
                    display = True

                if self.checkAll.isChecked():
                    display = True

                if display is True:
                    self.displayEvent(event)

                await op.sleep()

        except asyncio.CancelledError as err:
            return
        except Exception as err:
            log.debug('Unknown error ocurred while reading ipfs log')

    def onClose(self):
        self.tskLog.cancel()
        return True




import os, os.path
import sys
import asyncio
import tempfile
import functools
import re
import json

from galacteek import log

async def shell(arg):
    p = await asyncio.create_subprocess_shell((arg),
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE)

    stdout, stderr = await p.communicate()
    return stdout.decode()

async def ipfsConfig(binPath, param, value):
    return await shell("{0} config '{1}' '{2}'".format(
        binPath, param, value))

async def ipfsConfigJson(binPath, param, value):
    return await shell('{0} config --json {1} {2}'.format(
        binPath, param, json.dumps(value)))

async def ipfsConfigGetJson(binPath, param):
    return await shell('{0} config --json "{1}"'.format(binPath, param))

class IPFSDProtocol(asyncio.SubprocessProtocol):
    """
    IPFS daemon process protocol

    This handles output from the IPFS daemon
    """

    def __init__(self, loop, exitFuture, startedFuture):
        self._loop = loop
        self._output = bytearray()
        self.eventStarted = asyncio.Event()
        self.exitFuture = exitFuture
        self.startedFuture = startedFuture
        self.apiStarted = False
        self.gatewayStarted = False
        self.swarmStarted = False
        self.daemonReady = False
        self.errAlreadyRunning = False

    @property
    def loop(self):
        return self._loop

    @property
    def output(self):
        return self._output

    def pipe_data_received(self, fd, data):
        try:
            msg = data.decode().strip()
        except:
            return

        # The output we expect might be different in earlier versions
        # i've mainly tested with go-ipfs > 0.4.11
        # TODO: implement ipfs process supervisor independent of daemon output

        for line in msg.split('\n'):
            if re.search('Error: ipfs daemon is running', line):
                self.errAlreadyRunning = True
            if re.search('Gateway.*server listening on', line):
                self.gatewayStarted = True
            if re.search('Swarm listening on', line):
                self.swarmStarted = True
            if re.search('API server listening', line):
                self.apiStarted = True
            if re.search('Daemon is ready', line):
                self.daemonReady = True
            if re.search('Error:', line):
                pass

        if self.daemonReady is True:
            if not self.startedFuture.done():
                self.startedFuture.set_result(True)
            self.eventStarted.set()

        self._output.extend(data)

    def process_exited(self):
        self.exitFuture.set_result(True)

DEFAULT_APIPORT = 5001
DEFAULT_SWARMPORT = 4001
DEFAULT_GWPORT = 8080

class AsyncIPFSDaemon(object):
    """
    Async IPFS daemon subprocess wrapper

    :param str repopath: IPFS repository path
    :param int apiport: API port number to listen on
    :param int swarmport: Swarm port number to listen on
    :param int gatewayport: HTTP GW port number to listen on
    :param int swarmLowWater: min swarm connections
    :param int swarmHighWater: max swarm connections
    :param bool pubsubEnable: enable pubsub
    :param bool noBootstrap: empty bootstrap
    """
    def __init__(self, repopath, goIpfsPath='ipfs',
            apiport=DEFAULT_APIPORT,
            swarmport=DEFAULT_SWARMPORT,
            gatewayport=DEFAULT_GWPORT, initRepo=True,
            swarmLowWater=10, swarmHighWater=20,
            pubsubEnable=False, noBootstrap=False, corsEnable=True,
            p2pStreams=False, migrateRepo=False,
            storageMax=20, debug=False, loop=None):

        self.loop = loop if loop else asyncio.get_event_loop()
        self.exitFuture = asyncio.Future(loop=self.loop)
        self.startedFuture = asyncio.Future(loop=self.loop)
        self.repopath = repopath
        self.goIpfsPath = goIpfsPath
        self.apiport = apiport
        self.gatewayport = gatewayport
        self.swarmport = swarmport
        self.swarmLowWater = swarmLowWater
        self.swarmHighWater = swarmHighWater
        self.storageMax = storageMax
        self.initRepo = initRepo
        self.pubsubEnable = pubsubEnable
        self.corsEnable = corsEnable
        self.p2pStreams = p2pStreams
        self.noBootstrap = noBootstrap
        self.migrateRepo = migrateRepo
        self.debug = debug

    async def start(self):
        # Set the IPFS_PATH environment variable
        os.putenv('IPFS_PATH', self.repopath)
        if not os.path.isdir(self.repopath):
            os.mkdir(self.repopath)

        if not os.path.exists(os.path.join(self.repopath, 'config')) or \
            not os.path.isdir(os.path.join(self.repopath, 'datastore')):
            # Pretty sure this is an empty repository path
            initOutput = await shell('ipfs init')

        # Change the addresses/ports we listen on
        await ipfsConfig(self.goIpfsPath, 'Addresses.API',
                '/ip4/127.0.0.1/tcp/{0}'.format(self.apiport))
        await ipfsConfig(self.goIpfsPath, 'Addresses.Gateway',
                '/ip4/127.0.0.1/tcp/{0}'.format(self.gatewayport))

        swarmAddrs = [
            "/ip4/0.0.0.0/tcp/{swarmport}".format(swarmport=self.swarmport),
            "/ip6/::/tcp/{swarmport}".format(swarmport=self.swarmport)
        ]

        await ipfsConfigJson(self.goIpfsPath, 'Addresses.Swarm',
                json.dumps(swarmAddrs))

        # Swarm connection manager parameters
        await ipfsConfigJson(self.goIpfsPath, 'Swarm.ConnMgr.LowWater',
                self.swarmLowWater)
        await ipfsConfigJson(self.goIpfsPath, 'Swarm.ConnMgr.HighWater',
                self.swarmHighWater)

        # Maximum storage
        await ipfsConfig(self.goIpfsPath, 'Datastore.StorageMax',
                '{0}GB'.format(self.storageMax))

        # P2P
        if self.p2pStreams:
            await ipfsConfigJson(self.goIpfsPath,
                    'Experimental.Libp2pStreamMounting',
                    'true'
            )

        # CORS
        if self.corsEnable:
            # Setup the CORS headers, only allowing the gateway's origin
            await ipfsConfigJson(self.goIpfsPath,
                    'API.HTTPHeaders.Access-Control-Allow-Credentials',
                    '["true"]')
            await ipfsConfigJson(self.goIpfsPath,
                    'API.HTTPHeaders.Access-Control-Allow-Origin',
                    '["http://localhost:{0}"]'.format(self.gatewayport))

        if self.noBootstrap:
            await ipfsConfigJson(self.goIpfsPath, 'Bootstrap', '[]')

        args = [self.goIpfsPath, 'daemon']

        if self.pubsubEnable:
            args.append('--enable-pubsub-experiment')

        if self.migrateRepo:
            args.append('--migrate')

        f = self.loop.subprocess_exec(
                lambda: IPFSDProtocol(self.loop, self.exitFuture,
                    self.startedFuture),
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE)
        self.transport, self.proto = await f
        return True

    def stop(self):
        try:
            self.transport.terminate()
            return True
        except Exception as e:
            self.terminateException = e
            return False



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/hashmarksmgrnetwork.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_NetworkHashmarksViewForm(object):
    def setupUi(self, NetworkHashmarksViewForm):
        NetworkHashmarksViewForm.setObjectName("NetworkHashmarksViewForm")
        NetworkHashmarksViewForm.resize(541, 374)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(NetworkHashmarksViewForm)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.expandButton = QtWidgets.QToolButton(NetworkHashmarksViewForm)
        self.expandButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/expand.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.expandButton.setIcon(icon)
        self.expandButton.setCheckable(True)
        self.expandButton.setObjectName("expandButton")
        self.horizontalLayout.addWidget(self.expandButton)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.search = QtWidgets.QLineEdit(NetworkHashmarksViewForm)
        self.search.setObjectName("search")
        self.horizontalLayout.addWidget(self.search)
        self.searchButton = QtWidgets.QPushButton(NetworkHashmarksViewForm)
        self.searchButton.setObjectName("searchButton")
        self.horizontalLayout.addWidget(self.searchButton)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.treeNetMarks = QtWidgets.QTreeView(NetworkHashmarksViewForm)
        self.treeNetMarks.setObjectName("treeNetMarks")
        self.verticalLayout.addWidget(self.treeNetMarks)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(NetworkHashmarksViewForm)
        QtCore.QMetaObject.connectSlotsByName(NetworkHashmarksViewForm)

    def retranslateUi(self, NetworkHashmarksViewForm):
        _translate = QtCore.QCoreApplication.translate
        NetworkHashmarksViewForm.setWindowTitle(_translate("NetworkHashmarksViewForm", "Form"))
        self.searchButton.setText(_translate("NetworkHashmarksViewForm", "Search"))

from . import galacteek_rc



from multiaddr import Multiaddr

def multiAddrTcp4(maddr):
    ipaddr, port = None, 0

    try:
        multi = Multiaddr(maddr)
    except:
        return ipaddr, port

    for proto in multi.protocols():
        if proto.name == 'ip4':
            ipaddr = multi.value_for_protocol(proto.code)
        if proto.name == 'tcp':
            port = int(multi.value_for_protocol(proto.code))

    return ipaddr, port



import asyncio
import os.path

from PyQt5.QtWidgets import QTreeView, QMenu, QHeaderView, QPushButton
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QObject, QCoreApplication

from galacteek import asyncify, ensure
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.dag import DAGQuery
from galacteek.ipfs.cidhelpers import *
from galacteek.ipfs.ipfsops import *

from . import ui_peersmgr
from .widgets import *
from .modelhelpers import *
from .helpers import *
from .dialogs import *

def iUsername():
    return QCoreApplication.translate('PeersManager', 'Username')

def iPeerId():
    return QCoreApplication.translate('PeersManager', 'Peer ID')

def iPingAvg():
    return QCoreApplication.translate('PeersManager', 'Ping Average (ms)')

def iInvalidDagCID(dagCid):
    return QCoreApplication.translate('PeersManager',
            'Invalid DAG CID: {0}').format(dagCid)

class PeersModel(QStandardItemModel):
    pass

class PeersTracker(QObject):
    def __init__(self, ctx):
        super().__init__()
        self.model = PeersModel()
        self.model.setHorizontalHeaderLabels([
            iUsername(),
            iPeerId(),
            iPingAvg(),
            ''
        ])
        self.ctx = ctx
        self.ctx.peers.changed.connect(self.onPeersChange)
        self.ctx.peers.peerAdded.connect(self.onPeerAdded)
        self.ctx.peers.peerModified.connect(self.onPeerModified)
        self.ctx.peers.peerLogout.connect(self.onPeerLogout)
        self._peersRows = {}

    @property
    def modelRoot(self):
        return self.model.invisibleRootItem()

    @property
    def peersRows(self):
        return self._peersRows

    def onPeerLogout(self, peerId):
        ret = modelSearch(self.model,
                parent=self.modelRoot.index(),
                search=peerId, delete=True)
        if peerId in self._peersRows:
            del self._peersRows[peerId]

    def onPeerModified(self, peerId):
        peerCtx = self.ctx.peers.getByPeerId(peerId)
        if not peerCtx:
            return

        row = self._peersRows.get(peerId, None)
        if row:
            row[0].setText(peerCtx.ident.username)

    def onPeerAdded(self, peerId):
        peerCtx = self.ctx.peers.getByPeerId(peerId)
        if not peerCtx:
            return

        self.addPeerFromEntry(peerId, peerCtx)

    def addPeerFromEntry(self, peerId, peerCtx):
        row = [
            UneditableItem(peerCtx.ident.username),
            UneditableItem(peerId),
            UneditableItem(str(peerCtx.pingavg)),
            UneditableItem(''),
        ]
        row2 = [
            UneditableItem('Joined: {0}'.format(peerCtx.ident.dateCreated)),
            UneditableItem(peerId),
        ]

        self.modelRoot.appendRow(row)

        self._peersRows[peerId] = row
        root = row[0]
        root.appendRow(row2)

    def onPeersChange(self):
        pass

class PeersManager(GalacteekTab):
    def __init__(self, gWindow, peersTracker, **kw):
        super().__init__(gWindow, **kw)

        self.peersTracker = peersTracker
        self.ui = ui_peersmgr.Ui_PeersManager()
        self.ui.setupUi(self)

        self.ui.tree.setModel(self.peersTracker.model)
        self.ui.tree.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.ui.tree.header().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self.ui.tree.header().setSectionResizeMode(2, QHeaderView.ResizeToContents)

        self.ui.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.tree.customContextMenuRequested.connect(self.onContextMenu)
        self.ui.tree.doubleClicked.connect(self.onDoubleClick)

        self.ui.search.returnPressed.connect(self.onSearch)

        self.app.ipfsCtx.peers.changed.connect(self.onPeersChange)
        self.app.ipfsCtx.peers.peerAdded.connect(self.onPeerAdded)

    @property
    def model(self):
        return self.peersTracker.model

    def onSearch(self):
        search = self.ui.search.text()
        if len(search) > 0:
            self.ui.tree.keyboardSearch(search)

    def onPeersChange(self):
        self.ui.peersCountLabel.setText(str(self.app.ipfsCtx.peers.peersCount))

    def onPeerAdded(self, peerId):
        def openPeer(id):
            peerCtx = self.peersTracker.ctx.peers.getByPeerId(id)
            ensure(self.explorePeerHome(peerCtx))

        def followPeer(id):
            peerCtx = self.peersTracker.ctx.peers.getByPeerId(id)

            runDialog(AddFeedDialog, self.app.marksLocal,
                    joinIpns(peerCtx.ident.dagIpns),
                    feedName=peerCtx.ident.username)

        for peerId, row in self.peersTracker.peersRows.items():
            idx = self.model.indexFromItem(row[3])
            if self.ui.tree.indexWidget(idx) != None:
                continue
            btnEx = QPushButton('Homepage')
            btnEx.setFixedSize(100, 20)
            btnEx.clicked.connect(lambda: openPeer(peerId))

            self.ui.tree.setIndexWidget(idx, btnEx)

    def onContextMenu(self, point):
        idx = self.tree.indexAt(point)
        if not idx.isValid():
            return

        idxPeerId = self.model.sibling(idx.row(), 1, idx)
        peerId = self.model.data(idxPeerId)

        menu = QMenu()
        menu.exec(self.tree.mapToGlobal(point))

    def onDoubleClick(self, idx):
        idxPeerId = self.model.sibling(idx.row(), 1, idx)
        peerId = self.model.data(idxPeerId)
        peerCtx = self.peersTracker.ctx.peers.getByPeerId(peerId)

        ensure(self.explorePeerHome(peerCtx))

    @ipfsOp
    async def explorePeerHome(self, op, peerCtx):
        identMsg = peerCtx.ident

        dagCid = identMsg.dagCid
        if not cidValid(dagCid):
            log.debug('invalid DAG CID: {}'.format(dagCid))
            return messageBox(iInvalidDagCID(dagCid))

        self.gWindow.addBrowserTab().browseFsPath(
                os.path.join(joinIpns(identMsg.dagIpns), 'index.html'))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/files.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_FileManagerForm(object):
    def setupUi(self, FileManagerForm):
        FileManagerForm.setObjectName("FileManagerForm")
        FileManagerForm.resize(618, 349)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(FileManagerForm)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.pathSelector = QtWidgets.QComboBox(FileManagerForm)
        self.pathSelector.setObjectName("pathSelector")
        self.horizontalLayout_2.addWidget(self.pathSelector)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.addFileButton = QtWidgets.QToolButton(FileManagerForm)
        self.addFileButton.setMaximumSize(QtCore.QSize(32, 32))
        self.addFileButton.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/add-file.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.addFileButton.setIcon(icon)
        self.addFileButton.setIconSize(QtCore.QSize(24, 24))
        self.addFileButton.setObjectName("addFileButton")
        self.horizontalLayout_2.addWidget(self.addFileButton)
        self.addDirectoryButton = QtWidgets.QToolButton(FileManagerForm)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/share/icons/add-folder.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.addDirectoryButton.setIcon(icon1)
        self.addDirectoryButton.setIconSize(QtCore.QSize(24, 24))
        self.addDirectoryButton.setObjectName("addDirectoryButton")
        self.horizontalLayout_2.addWidget(self.addDirectoryButton)
        self.refreshButton = QtWidgets.QToolButton(FileManagerForm)
        self.refreshButton.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/share/icons/refresh.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.refreshButton.setIcon(icon2)
        self.refreshButton.setIconSize(QtCore.QSize(24, 24))
        self.refreshButton.setObjectName("refreshButton")
        self.horizontalLayout_2.addWidget(self.refreshButton)
        self.horizontalLayout.addLayout(self.horizontalLayout_2)
        spacerItem1 = QtWidgets.QSpacerItem(10, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.fileManagerButton = QtWidgets.QToolButton(FileManagerForm)
        self.fileManagerButton.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/share/icons/file-manager.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.fileManagerButton.setIcon(icon3)
        self.fileManagerButton.setIconSize(QtCore.QSize(24, 24))
        self.fileManagerButton.setObjectName("fileManagerButton")
        self.horizontalLayout.addWidget(self.fileManagerButton)
        self.statusLabel = QtWidgets.QLabel(FileManagerForm)
        self.statusLabel.setMinimumSize(QtCore.QSize(200, 0))
        self.statusLabel.setText("")
        self.statusLabel.setObjectName("statusLabel")
        self.horizontalLayout.addWidget(self.statusLabel)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(FileManagerForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setMaximumSize(QtCore.QSize(100, 16777215))
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.searchFiles = QtWidgets.QLineEdit(FileManagerForm)
        self.searchFiles.setMaximumSize(QtCore.QSize(500, 16777215))
        self.searchFiles.setObjectName("searchFiles")
        self.horizontalLayout_3.addWidget(self.searchFiles)
        self.horizontalLayout.addLayout(self.horizontalLayout_3)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.hLayoutFilesView = QtWidgets.QHBoxLayout()
        self.hLayoutFilesView.setObjectName("hLayoutFilesView")
        self.treeFiles = QtWidgets.QTreeView(FileManagerForm)
        self.treeFiles.setObjectName("treeFiles")
        self.treeFiles.header().setCascadingSectionResizes(True)
        self.treeFiles.header().setDefaultSectionSize(140)
        self.treeFiles.header().setStretchLastSection(True)
        self.hLayoutFilesView.addWidget(self.treeFiles)
        self.hLayoutBrowser = QtWidgets.QHBoxLayout()
        self.hLayoutBrowser.setObjectName("hLayoutBrowser")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.fileManagerSwitch = QtWidgets.QPushButton(FileManagerForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fileManagerSwitch.sizePolicy().hasHeightForWidth())
        self.fileManagerSwitch.setSizePolicy(sizePolicy)
        self.fileManagerSwitch.setMaximumSize(QtCore.QSize(128, 16777215))
        self.fileManagerSwitch.setText("")
        self.fileManagerSwitch.setIcon(icon3)
        self.fileManagerSwitch.setObjectName("fileManagerSwitch")
        self.verticalLayout_4.addWidget(self.fileManagerSwitch)
        self.hLayoutBrowser.addLayout(self.verticalLayout_4)
        self.hLayoutFilesView.addLayout(self.hLayoutBrowser)
        self.verticalLayout.addLayout(self.hLayoutFilesView)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(FileManagerForm)
        QtCore.QMetaObject.connectSlotsByName(FileManagerForm)

    def retranslateUi(self, FileManagerForm):
        _translate = QtCore.QCoreApplication.translate
        FileManagerForm.setWindowTitle(_translate("FileManagerForm", "Form"))
        self.addFileButton.setToolTip(_translate("FileManagerForm", "<html><head/><body><p>Add Files</p></body></html>"))
        self.addDirectoryButton.setToolTip(_translate("FileManagerForm", "<html><head/><body><p>Add directory</p></body></html>"))
        self.addDirectoryButton.setText(_translate("FileManagerForm", "Add directory"))
        self.refreshButton.setToolTip(_translate("FileManagerForm", "<html><head/><body><p>Refresh</p></body></html>"))
        self.fileManagerButton.setToolTip(_translate("FileManagerForm", "<html><head/><body><p>Open local file manager</p></body></html>"))
        self.label.setText(_translate("FileManagerForm", "Search"))
        self.fileManagerSwitch.setToolTip(_translate("FileManagerForm", "<html><head/><body><p>Open local file manager</p></body></html>"))

from . import galacteek_rc



countryList = [
  {'name': 'Afghanistan', 'code': 'AF'},
  {'name': 'Åland Islands', 'code': 'AX'},
  {'name': 'Albania', 'code': 'AL'},
  {'name': 'Algeria', 'code': 'DZ'},
  {'name': 'American Samoa', 'code': 'AS'},
  {'name': 'AndorrA', 'code': 'AD'},
  {'name': 'Angola', 'code': 'AO'},
  {'name': 'Anguilla', 'code': 'AI'},
  {'name': 'Antarctica', 'code': 'AQ'},
  {'name': 'Antigua and Barbuda', 'code': 'AG'},
  {'name': 'Argentina', 'code': 'AR'},
  {'name': 'Armenia', 'code': 'AM'},
  {'name': 'Aruba', 'code': 'AW'},
  {'name': 'Australia', 'code': 'AU'},
  {'name': 'Austria', 'code': 'AT'},
  {'name': 'Azerbaijan', 'code': 'AZ'},
  {'name': 'Bahamas', 'code': 'BS'},
  {'name': 'Bahrain', 'code': 'BH'},
  {'name': 'Bangladesh', 'code': 'BD'},
  {'name': 'Barbados', 'code': 'BB'},
  {'name': 'Belarus', 'code': 'BY'},
  {'name': 'Belgium', 'code': 'BE'},
  {'name': 'Belize', 'code': 'BZ'},
  {'name': 'Benin', 'code': 'BJ'},
  {'name': 'Bermuda', 'code': 'BM'},
  {'name': 'Bhutan', 'code': 'BT'},
  {'name': 'Bolivia', 'code': 'BO'},
  {'name': 'Bosnia and Herzegovina', 'code': 'BA'},
  {'name': 'Botswana', 'code': 'BW'},
  {'name': 'Bouvet Island', 'code': 'BV'},
  {'name': 'Brazil', 'code': 'BR'},
  {'name': 'British Indian Ocean Territory', 'code': 'IO'},
  {'name': 'Brunei Darussalam', 'code': 'BN'},
  {'name': 'Bulgaria', 'code': 'BG'},
  {'name': 'Burkina Faso', 'code': 'BF'},
  {'name': 'Burundi', 'code': 'BI'},
  {'name': 'Cambodia', 'code': 'KH'},
  {'name': 'Cameroon', 'code': 'CM'},
  {'name': 'Canada', 'code': 'CA'},
  {'name': 'Cape Verde', 'code': 'CV'},
  {'name': 'Cayman Islands', 'code': 'KY'},
  {'name': 'Central African Republic', 'code': 'CF'},
  {'name': 'Chad', 'code': 'TD'},
  {'name': 'Chile', 'code': 'CL'},
  {'name': 'China', 'code': 'CN'},
  {'name': 'Christmas Island', 'code': 'CX'},
  {'name': 'Cocos (Keeling) Islands', 'code': 'CC'},
  {'name': 'Colombia', 'code': 'CO'},
  {'name': 'Comoros', 'code': 'KM'},
  {'name': 'Congo', 'code': 'CG'},
  {'name': 'Congo, The Democratic Republic of the', 'code': 'CD'},
  {'name': 'Cook Islands', 'code': 'CK'},
  {'name': 'Costa Rica', 'code': 'CR'},
  {'name': 'Cote D\'Ivoire', 'code': 'CI'},
  {'name': 'Croatia', 'code': 'HR'},
  {'name': 'Cuba', 'code': 'CU'},
  {'name': 'Cyprus', 'code': 'CY'},
  {'name': 'Czech Republic', 'code': 'CZ'},
  {'name': 'Denmark', 'code': 'DK'},
  {'name': 'Djibouti', 'code': 'DJ'},
  {'name': 'Dominica', 'code': 'DM'},
  {'name': 'Dominican Republic', 'code': 'DO'},
  {'name': 'Ecuador', 'code': 'EC'},
  {'name': 'Egypt', 'code': 'EG'},
  {'name': 'El Salvador', 'code': 'SV'},
  {'name': 'Equatorial Guinea', 'code': 'GQ'},
  {'name': 'Eritrea', 'code': 'ER'},
  {'name': 'Estonia', 'code': 'EE'},
  {'name': 'Ethiopia', 'code': 'ET'},
  {'name': 'Falkland Islands (Malvinas)', 'code': 'FK'},
  {'name': 'Faroe Islands', 'code': 'FO'},
  {'name': 'Fiji', 'code': 'FJ'},
  {'name': 'Finland', 'code': 'FI'},
  {'name': 'France', 'code': 'FR'},
  {'name': 'French Guiana', 'code': 'GF'},
  {'name': 'French Polynesia', 'code': 'PF'},
  {'name': 'French Southern Territories', 'code': 'TF'},
  {'name': 'Gabon', 'code': 'GA'},
  {'name': 'Gambia', 'code': 'GM'},
  {'name': 'Georgia', 'code': 'GE'},
  {'name': 'Germany', 'code': 'DE'},
  {'name': 'Ghana', 'code': 'GH'},
  {'name': 'Gibraltar', 'code': 'GI'},
  {'name': 'Greece', 'code': 'GR'},
  {'name': 'Greenland', 'code': 'GL'},
  {'name': 'Grenada', 'code': 'GD'},
  {'name': 'Guadeloupe', 'code': 'GP'},
  {'name': 'Guam', 'code': 'GU'},
  {'name': 'Guatemala', 'code': 'GT'},
  {'name': 'Guernsey', 'code': 'GG'},
  {'name': 'Guinea', 'code': 'GN'},
  {'name': 'Guinea-Bissau', 'code': 'GW'},
  {'name': 'Guyana', 'code': 'GY'},
  {'name': 'Haiti', 'code': 'HT'},
  {'name': 'Heard Island and Mcdonald Islands', 'code': 'HM'},
  {'name': 'Holy See (Vatican City State)', 'code': 'VA'},
  {'name': 'Honduras', 'code': 'HN'},
  {'name': 'Hong Kong', 'code': 'HK'},
  {'name': 'Hungary', 'code': 'HU'},
  {'name': 'Iceland', 'code': 'IS'},
  {'name': 'India', 'code': 'IN'},
  {'name': 'Indonesia', 'code': 'ID'},
  {'name': 'Iran, Islamic Republic Of', 'code': 'IR'},
  {'name': 'Iraq', 'code': 'IQ'},
  {'name': 'Ireland', 'code': 'IE'},
  {'name': 'Isle of Man', 'code': 'IM'},
  {'name': 'Israel', 'code': 'IL'},
  {'name': 'Italy', 'code': 'IT'},
  {'name': 'Jamaica', 'code': 'JM'},
  {'name': 'Japan', 'code': 'JP'},
  {'name': 'Jersey', 'code': 'JE'},
  {'name': 'Jordan', 'code': 'JO'},
  {'name': 'Kazakhstan', 'code': 'KZ'},
  {'name': 'Kenya', 'code': 'KE'},
  {'name': 'Kiribati', 'code': 'KI'},
  {'name': 'Korea, Democratic People\'S Republic of', 'code': 'KP'},
  {'name': 'Korea, Republic of', 'code': 'KR'},
  {'name': 'Kuwait', 'code': 'KW'},
  {'name': 'Kyrgyzstan', 'code': 'KG'},
  {'name': 'Lao People\'S Democratic Republic', 'code': 'LA'},
  {'name': 'Latvia', 'code': 'LV'},
  {'name': 'Lebanon', 'code': 'LB'},
  {'name': 'Lesotho', 'code': 'LS'},
  {'name': 'Liberia', 'code': 'LR'},
  {'name': 'Libyan Arab Jamahiriya', 'code': 'LY'},
  {'name': 'Liechtenstein', 'code': 'LI'},
  {'name': 'Lithuania', 'code': 'LT'},
  {'name': 'Luxembourg', 'code': 'LU'},
  {'name': 'Macao', 'code': 'MO'},
  {'name': 'Macedonia, The Former Yugoslav Republic of', 'code': 'MK'},
  {'name': 'Madagascar', 'code': 'MG'},
  {'name': 'Malawi', 'code': 'MW'},
  {'name': 'Malaysia', 'code': 'MY'},
  {'name': 'Maldives', 'code': 'MV'},
  {'name': 'Mali', 'code': 'ML'},
  {'name': 'Malta', 'code': 'MT'},
  {'name': 'Marshall Islands', 'code': 'MH'},
  {'name': 'Martinique', 'code': 'MQ'},
  {'name': 'Mauritania', 'code': 'MR'},
  {'name': 'Mauritius', 'code': 'MU'},
  {'name': 'Mayotte', 'code': 'YT'},
  {'name': 'Mexico', 'code': 'MX'},
  {'name': 'Micronesia, Federated States of', 'code': 'FM'},
  {'name': 'Moldova, Republic of', 'code': 'MD'},
  {'name': 'Monaco', 'code': 'MC'},
  {'name': 'Mongolia', 'code': 'MN'},
  {'name': 'Montserrat', 'code': 'MS'},
  {'name': 'Morocco', 'code': 'MA'},
  {'name': 'Mozambique', 'code': 'MZ'},
  {'name': 'Myanmar', 'code': 'MM'},
  {'name': 'Namibia', 'code': 'NA'},
  {'name': 'Nauru', 'code': 'NR'},
  {'name': 'Nepal', 'code': 'NP'},
  {'name': 'Netherlands', 'code': 'NL'},
  {'name': 'Netherlands Antilles', 'code': 'AN'},
  {'name': 'New Caledonia', 'code': 'NC'},
  {'name': 'New Zealand', 'code': 'NZ'},
  {'name': 'Nicaragua', 'code': 'NI'},
  {'name': 'Niger', 'code': 'NE'},
  {'name': 'Nigeria', 'code': 'NG'},
  {'name': 'Niue', 'code': 'NU'},
  {'name': 'Norfolk Island', 'code': 'NF'},
  {'name': 'Northern Mariana Islands', 'code': 'MP'},
  {'name': 'Norway', 'code': 'NO'},
  {'name': 'Oman', 'code': 'OM'},
  {'name': 'Pakistan', 'code': 'PK'},
  {'name': 'Palau', 'code': 'PW'},
  {'name': 'Palestinian Territory, Occupied', 'code': 'PS'},
  {'name': 'Panama', 'code': 'PA'},
  {'name': 'Papua New Guinea', 'code': 'PG'},
  {'name': 'Paraguay', 'code': 'PY'},
  {'name': 'Peru', 'code': 'PE'},
  {'name': 'Philippines', 'code': 'PH'},
  {'name': 'Pitcairn', 'code': 'PN'},
  {'name': 'Poland', 'code': 'PL'},
  {'name': 'Portugal', 'code': 'PT'},
  {'name': 'Puerto Rico', 'code': 'PR'},
  {'name': 'Qatar', 'code': 'QA'},
  {'name': 'Reunion', 'code': 'RE'},
  {'name': 'Romania', 'code': 'RO'},
  {'name': 'Russian Federation', 'code': 'RU'},
  {'name': 'RWANDA', 'code': 'RW'},
  {'name': 'Saint Helena', 'code': 'SH'},
  {'name': 'Saint Kitts and Nevis', 'code': 'KN'},
  {'name': 'Saint Lucia', 'code': 'LC'},
  {'name': 'Saint Pierre and Miquelon', 'code': 'PM'},
  {'name': 'Saint Vincent and the Grenadines', 'code': 'VC'},
  {'name': 'Samoa', 'code': 'WS'},
  {'name': 'San Marino', 'code': 'SM'},
  {'name': 'Sao Tome and Principe', 'code': 'ST'},
  {'name': 'Saudi Arabia', 'code': 'SA'},
  {'name': 'Senegal', 'code': 'SN'},
  {'name': 'Serbia and Montenegro', 'code': 'CS'},
  {'name': 'Seychelles', 'code': 'SC'},
  {'name': 'Sierra Leone', 'code': 'SL'},
  {'name': 'Singapore', 'code': 'SG'},
  {'name': 'Slovakia', 'code': 'SK'},
  {'name': 'Slovenia', 'code': 'SI'},
  {'name': 'Solomon Islands', 'code': 'SB'},
  {'name': 'Somalia', 'code': 'SO'},
  {'name': 'South Africa', 'code': 'ZA'},
  {'name': 'South Georgia and the South Sandwich Islands', 'code': 'GS'},
  {'name': 'Spain', 'code': 'ES'},
  {'name': 'Sri Lanka', 'code': 'LK'},
  {'name': 'Sudan', 'code': 'SD'},
  {'name': 'Suriname', 'code': 'SR'},
  {'name': 'Svalbard and Jan Mayen', 'code': 'SJ'},
  {'name': 'Swaziland', 'code': 'SZ'},
  {'name': 'Sweden', 'code': 'SE'},
  {'name': 'Switzerland', 'code': 'CH'},
  {'name': 'Syrian Arab Republic', 'code': 'SY'},
  {'name': 'Taiwan, Province of China', 'code': 'TW'},
  {'name': 'Tajikistan', 'code': 'TJ'},
  {'name': 'Tanzania, United Republic of', 'code': 'TZ'},
  {'name': 'Thailand', 'code': 'TH'},
  {'name': 'Timor-Leste', 'code': 'TL'},
  {'name': 'Togo', 'code': 'TG'},
  {'name': 'Tokelau', 'code': 'TK'},
  {'name': 'Tonga', 'code': 'TO'},
  {'name': 'Trinidad and Tobago', 'code': 'TT'},
  {'name': 'Tunisia', 'code': 'TN'},
  {'name': 'Turkey', 'code': 'TR'},
  {'name': 'Turkmenistan', 'code': 'TM'},
  {'name': 'Turks and Caicos Islands', 'code': 'TC'},
  {'name': 'Tuvalu', 'code': 'TV'},
  {'name': 'Uganda', 'code': 'UG'},
  {'name': 'Ukraine', 'code': 'UA'},
  {'name': 'United Arab Emirates', 'code': 'AE'},
  {'name': 'United Kingdom', 'code': 'GB'},
  {'name': 'United States', 'code': 'US'},
  {'name': 'United States Minor Outlying Islands', 'code': 'UM'},
  {'name': 'Uruguay', 'code': 'UY'},
  {'name': 'Uzbekistan', 'code': 'UZ'},
  {'name': 'Vanuatu', 'code': 'VU'},
  {'name': 'Venezuela', 'code': 'VE'},
  {'name': 'Viet Nam', 'code': 'VN'},
  {'name': 'Virgin Islands, British', 'code': 'VG'},
  {'name': 'Virgin Islands, U.S.', 'code': 'VI'},
  {'name': 'Wallis and Futuna', 'code': 'WF'},
  {'name': 'Western Sahara', 'code': 'EH'},
  {'name': 'Yemen', 'code': 'YE'},
  {'name': 'Zambia', 'code': 'ZM'},
  {'name': 'Zimbabwe', 'code': 'ZW'}
]




import sys
import time
import os.path
import re

from PyQt5.QtWidgets import (QWidget, QFrame, QApplication, QMainWindow,
        QDialog, QLabel, QPushButton, QVBoxLayout, QAction, QStyle,
        QMenu, QTabWidget, QInputDialog, QMessageBox, QToolButton, QFileDialog)

from PyQt5.QtPrintSupport import *

from PyQt5.QtCore import (QUrl, QIODevice, Qt, QCoreApplication, QObject,
    pyqtSignal, QMutex, QFile)
from PyQt5 import QtWebEngineWidgets, QtWebEngine, QtWebEngineCore
from PyQt5.QtWebEngineWidgets import (QWebEngineDownloadItem, QWebEngineScript,
        QWebEngineSettings)
from PyQt5.QtWebEngineCore import QWebEngineUrlRequestInterceptor
from PyQt5.Qt import QByteArray
from PyQt5.QtGui import QClipboard, QPixmap, QIcon, QKeySequence

from yarl import URL

from galacteek import log
from galacteek.ipfs.wrappers import *

from . import ui_browsertab
from . import galacteek_rc
from .helpers import *
from .dialogs import *
from .hashmarks import *
from .i18n import *
from .widgets import *
from ..appsettings import *
from galacteek.ipfs import cidhelpers

SCHEME_DWEB = 'dweb'
SCHEME_FS = 'fs'
SCHEME_IPFS = 'ipfs'

# i18n
def iOpenInTab():
    return QCoreApplication.translate('BrowserTabForm', 'Open link in tab')
def iOpenWith():
    return QCoreApplication.translate('BrowserTabForm', 'Open with')
def iDownload():
    return QCoreApplication.translate('BrowserTabForm', 'Download')
def iPin():
    return QCoreApplication.translate('BrowserTabForm', 'PIN')
def iPinThisPage():
    return QCoreApplication.translate('BrowserTabForm', 'PIN (this page)')
def iPinRecursive():
    return QCoreApplication.translate('BrowserTabForm', 'PIN (recursive)')
def iFollow():
    return QCoreApplication.translate('BrowserTabForm',
        'Follow IPNS resource')

def iEnterIpfsCID():
    return QCoreApplication.translate('BrowserTabForm', 'Enter an IPFS CID')

def iBrowseHomePage():
    return QCoreApplication.translate('BrowserTabForm',
        'Go to home page')

def iBrowseIpfsCID():
    return QCoreApplication.translate('BrowserTabForm',
        'Browse IPFS resource (CID)')

def iBrowseIpfsMultipleCID():
    return QCoreApplication.translate('BrowserTabForm',
        'Browse multiple IPFS resources (CID)')

def iEnterIpfsCIDDialog():
    return QCoreApplication.translate('BrowserTabForm',
        'Load IPFS CID dialog')

def iFollowIpnsDialog():
    return QCoreApplication.translate('BrowserTabForm',
        'IPNS add feed dialog')

def iBrowseIpnsHash():
    return QCoreApplication.translate('BrowserTabForm',
        'Browse IPNS resource from hash/name')

def iEnterIpns():
    return QCoreApplication.translate('BrowserTabForm',
        'Enter an IPNS hash/name')

def iEnterIpnsDialog():
    return QCoreApplication.translate('BrowserTabForm',
        'Load IPNS key dialog')

def iHashmarked(path):
    return QCoreApplication.translate('BrowserTabForm',
        'Hashmarked {0}').format(path)

def iHashmarkTitleDialog():
    return QCoreApplication.translate('BrowserTabForm',
        'Hashmark title')

def iInvalidUrl(text):
    return QCoreApplication.translate('BrowserTabForm',
        'Invalid URL: {0}').format(text)

def iInvalidCID(text):
    return QCoreApplication.translate('BrowserTabForm',
        '{0} is an invalid IPFS CID (Content IDentifier)').format(text)

def fsPath(path):
    return '{0}:{1}'.format(SCHEME_IPFS, path)

def usesIpfsNs(url):
    return url.startswith('{0}:/'.format(SCHEME_IPFS))

class IPFSSchemeHandler(QtWebEngineCore.QWebEngineUrlSchemeHandler):
    def __init__(self, app, parent=None):
        QtWebEngineCore.QWebEngineUrlSchemeHandler.__init__(self, parent)
        self.app = app

    def requestStarted(self, request):
        url = request.requestUrl()
        scheme = url.scheme()
        path = url.path()

        log.debug('IPFS scheme handler request {url} {scheme} {path} {method}'.format(
            url=url.toString(), scheme=scheme, path=path,
            method=request.requestMethod()))

        def redirectIpfs(path):
            yUrl = URL(url.toString())

            if len(yUrl.parts) < 3:
                messageBox(iInvalidUrl())
                return None

            newUrl = self.app.subUrl(path)

            if url.hasFragment():
                newUrl.setFragment(url.fragment())
            if url.hasQuery():
                newUrl.setQuery(url.query())

            log.debug('IPFS scheme handler redirects to {redirect} {scheme} \
                    {valid}'.format(
                redirect=newUrl.toString(), scheme=newUrl.scheme(),
                valid=newUrl.isValid()))
            return request.redirect(newUrl)

        if scheme in [SCHEME_FS, SCHEME_IPFS, SCHEME_DWEB]:
            # Handle scheme:/{ipfs,ipns}/path

            if path.startswith('/ipfs/') or path.startswith('/ipns/'):
                return redirectIpfs(path)

class RequestInterceptor(QWebEngineUrlRequestInterceptor):
    def interceptRequest(self, info):
        url = info.requestUrl()

class WebView(QtWebEngineWidgets.QWebEngineView):
    def __init__(self, browserTab, enablePlugins=False, parent = None):
        super(QtWebEngineWidgets.QWebEngineView, self).__init__(parent = parent)

        self.mutex = QMutex()
        self.browserTab = browserTab

        self.webSettings = self.settings()
        self.webSettings.setAttribute(QWebEngineSettings.PluginsEnabled,
                enablePlugins)

    def contextMenuEvent(self, event):
        currentPage = self.page()
        contextMenuData = currentPage.contextMenuData()
        menu = self.page().createStandardContextMenu()
        actions = menu.actions()
        menu.addSeparator()

        act1 = menu.addAction(iOpenInTab(), lambda:
                self.openInTab(contextMenuData))
        act1 = menu.addAction(iDownload(), lambda:
                self.downloadLink(contextMenuData))
        ipfsMenu = QMenu('IPFS')
        ipfsMenu.setIcon(getIcon('ipfs-logo-128-black.png'))
        menu.addMenu(ipfsMenu)
        act = ipfsMenu.addAction(getIcon('pin.png'), iPin(), lambda:
                self.pinPage(contextMenuData))
        openWithMenu = QMenu(iOpenWith())
        openWithMenu.addAction('Media player', lambda:
                self.openWithMediaPlayer(contextMenuData))
        menu.addMenu(openWithMenu)

        menu.exec(event.globalPos())

    def openWithMediaPlayer(self, menudata):
        url = menudata.linkUrl()
        self.browserTab.gWindow.mediaPlayerQueue(url.path())

    def pinPage(self, menudata):
        url = menudata.linkUrl()
        path = url.path()
        self.browserTab.pinPath(path, recursive=False)

    def openInTab(self, menudata):
        url = menudata.linkUrl()
        tab = self.browserTab.gWindow.addBrowserTab()
        tab.enterUrl(url)

    def downloadLink(self, menudata):
        url = menudata.linkUrl()
        self.page().download(url, None)

    def createWindow(self, wtype):
        pass

class BrowserKeyFilter(QObject):
    hashmarkPressed = pyqtSignal()
    savePagePressed = pyqtSignal()

    def eventFilter(self,  obj,  event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()

            key = event.key()
            if modifiers & Qt.ControlModifier:
                if key == Qt.Key_B:
                    self.hashmarkPressed.emit()
                    return True
                if key == Qt.Key_S:
                    self.savePagePressed.emit()
                    return True
        return False

class BrowserTab(GalacteekTab):
    # signals
    ipfsPathVisited = pyqtSignal(str)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.ui = ui_browsertab.Ui_BrowserTabForm()
        self.ui.setupUi(self)

        # Install scheme handler early on
        self.webProfile = QtWebEngineWidgets.QWebEngineProfile.defaultProfile()
        self.installIpfsSchemeHandler()

        self.ui.webEngineView = WebView(self,
                enablePlugins=self.app.settingsMgr.ppApiPlugins)
        self.ui.vLayoutBrowser.addWidget(self.ui.webEngineView)

        self.webScripts = self.webProfile.scripts()
        self.installScripts()

        self.ui.webEngineView.urlChanged.connect(self.onUrlChanged)
        self.ui.webEngineView.loadFinished.connect(self.onLoadFinished)
        self.ui.webEngineView.iconChanged.connect(self.onIconChanged)
        self.ui.webEngineView.loadProgress.connect(self.onLoadProgress)

        self.ui.urlZone.returnPressed.connect(self.onUrlEdit)
        self.ui.backButton.clicked.connect(self.backButtonClicked)
        self.ui.backButton.setIcon(self.style().standardIcon(QStyle.SP_ArrowLeft))
        self.ui.forwardButton.clicked.connect(self.forwardButtonClicked)
        self.ui.forwardButton.setIcon(self.style().standardIcon(QStyle.SP_ArrowRight))
        self.ui.refreshButton.clicked.connect(self.refreshButtonClicked)
        self.ui.stopButton.clicked.connect(self.stopButtonClicked)
        self.ui.loadFromClipboardButton.clicked.connect(self.loadFromClipboardButtonClicked)
        self.ui.loadFromClipboardButton.setEnabled(self.app.clipTracker.hasIpfs)
        self.ui.hashmarkPageButton.clicked.connect(self.onHashmarkPage)

        # Setup the tool button for browsing IPFS content
        self.loadIpfsMenu = QMenu()
        self.loadIpfsCIDAction = QAction(getIconIpfsIce(),
                iBrowseIpfsCID(),self,
                shortcut=QKeySequence('Ctrl+l'),
                triggered=self.onLoadIpfsCID)
        self.loadIpfsMultipleCIDAction = QAction(getIconIpfsIce(),
                iBrowseIpfsMultipleCID(), self,
                triggered=self.onLoadIpfsMultipleCID)
        self.loadIpnsAction = QAction(getIconIpfsWhite(),
                iBrowseIpnsHash(),self,
                triggered=self.onLoadIpns)
        self.followIpnsAction = QAction(getIconIpfsWhite(),
                iFollow(),self,
                triggered=self.onFollowIpns)
        self.loadHomeAction = QAction(getIcon('go-home.png'),
                iBrowseHomePage(),self,
                triggered=self.onLoadHome)

        self.loadIpfsMenu.addAction(self.loadIpfsCIDAction)
        self.loadIpfsMenu.addAction(self.loadIpfsMultipleCIDAction)
        self.loadIpfsMenu.addAction(self.loadIpnsAction)
        self.loadIpfsMenu.addAction(self.followIpnsAction)
        self.loadIpfsMenu.addAction(self.loadHomeAction)

        self.ui.loadIpfsButton.setMenu(self.loadIpfsMenu)
        self.ui.loadIpfsButton.setPopupMode(QToolButton.MenuButtonPopup)
        self.ui.loadIpfsButton.clicked.connect(self.onLoadIpfsCID)

        self.ui.pinAllButton.setCheckable(True)
        self.ui.pinAllButton.setAutoRaise(True)
        self.ui.pinAllButton.setChecked(self.gWindow.pinAllGlobalChecked)
        self.ui.pinAllButton.toggled.connect(self.onToggledPinAll)

        # Prepare the pin combo box
        iconPin = getIcon('pin.png')
        self.ui.actionComboBox.insertItem(0, iPinThisPage())
        self.ui.actionComboBox.setItemIcon(0, iconPin)
        self.ui.actionComboBox.insertItem(1, iPinRecursive())
        self.ui.actionComboBox.setItemIcon(1, iconPin)
        self.ui.actionComboBox.activated.connect(self.actionComboClicked)

        # Event filter
        evfilter = BrowserKeyFilter(self)
        evfilter.hashmarkPressed.connect(self.onHashmarkPage)
        self.installEventFilter(evfilter)

        self.app.clipTracker.clipboardHasIpfs.connect(self.onClipboardIpfs)
        self.ipfsPathVisited.connect(self.onPathVisited)

        self.currentUrl = None
        self.currentIpfsResource = None
        self.currentPageTitle = None

    @property
    def webView(self):
        return self.ui.webEngineView

    @property
    def tabPage(self):
        return self.gWindow.ui.tabWidget.widget(self.tabPageIdx)

    @property
    def tabPageIdx(self):
        return self.gWindow.ui.tabWidget.indexOf(self)

    @property
    def gatewayAuthority(self):
        return self.app.gatewayAuthority

    @property
    def gatewayUrl(self):
        return self.app.gatewayUrl

    @property
    def pinAll(self):
        return self.ui.pinAllButton.isChecked()

    def installScripts(self):
        # Install the browserified js-ipfs API

        exSc = self.webScripts.findScript('js-ipfs-api')
        if self.app.settingsMgr.jsIpfsApi is True and exSc.isNull():
            jsFile = QFile(':/share/js/js-ipfs-api/index.js')
            if not jsFile.open(QFile.ReadOnly):
                return
            scriptJsIpfs = QWebEngineScript()
            scriptJsIpfs.setName('js-ipfs-api')
            scriptJsIpfs.setSourceCode(jsFile.readAll().data().decode('utf-8'))
            scriptJsIpfs.setWorldId(QWebEngineScript.MainWorld)
            scriptJsIpfs.setInjectionPoint(QWebEngineScript.DocumentCreation)
            scriptJsIpfs.setRunsOnSubFrames(True)
            self.webScripts.insert(scriptJsIpfs)

            # Install another script creating an IpfsApi instance
            # The API is available from Javascript as 'window.ipfs'

            script = QWebEngineScript()
            script.setSourceCode('\n'.join([
                "document.addEventListener('DOMContentLoaded', function () {",
                "window.ipfs = window.IpfsApi('{host}', '{port}');".format(
                    host=self.app.getIpfsConnectionParams().host,
                    port=self.app.getIpfsConnectionParams().apiPort),
                "})"]))
            script.setWorldId(QWebEngineScript.MainWorld)
            script.setInjectionPoint(QWebEngineScript.DocumentCreation)
            self.webScripts.insert(script)

    def installIpfsSchemeHandler(self):
        baFs   = QByteArray(b'fs')
        baIpfs = QByteArray(b'ipfs')
        baDweb = QByteArray(b'dweb')

        for scheme in [baFs, baIpfs, baDweb]:
            eHandler = self.webProfile.urlSchemeHandler(scheme)
            if not eHandler:
                self.webProfile.installUrlSchemeHandler(scheme,
                    self.app.ipfsSchemeHandler)

    def onPathVisited(self, path):
        # Called after a new IPFS object has been loaded in this tab
        self.app.task(self.tsVisitPath, path)

    @ipfsStatOp
    async def tsVisitPath(self, ipfsop, path, stat):
        # If automatic pinning is checked we pin the object
        if self.pinAll is True:
            self.pinPath(path, recursive=False, notify=False)

    def onClipboardIpfs(self, valid, cid, path):
        self.ui.loadFromClipboardButton.setEnabled(valid)

    def onToggledPinAll(self, checked):
        pass

    def onSavePage(self):
        page = self.webView.page()

        result = QFileDialog.getSaveFileName(None,
            'Select file',
            self.app.settingsMgr.getSetting(CFG_SECTION_BROWSER,
                CFG_KEY_DLPATH), '(*.*)')
        if not result:
            return

        page.save(result[0], QWebEngineDownloadItem.CompleteHtmlSaveFormat)

    def onHashmarkPage(self):
        if self.currentIpfsResource:
            addHashmark(self.app.marksLocal,
                    self.currentIpfsResource,
                    self.currentPageTitle,
                    stats=self.app.ipfsCtx.objectStats.get(
                        self.currentIpfsResource, {}))

    def onPinSuccess(self, f):
        self.app.systemTrayMessage('PIN', iPinSuccess(f.result()))

    def pinPath(self, path, recursive=True, notify=True):
        async def pinCoro(client, path):
            pinner = self.app.ipfsCtx.pinner
            onSuccess = None
            if notify is True:
                onSuccess = self.onPinSuccess
            await pinner.queue(path, recursive, onSuccess)

        self.app.ipfsTask(pinCoro, path)

    def printButtonClicked(self):
        # TODO: reinstate the printing button
        printer = QPrinter()
        dialog = QPrintDialog(printer, self)
        dialog.setModal(True)

        def success(ok):
            return ok

        if dialog.exec_() == QDialog.Accepted:
            currentPage = self.ui.webEngineView.page()
            currentPage.print(printer, success)

    def actionComboClicked(self, idx):
        if not self.currentIpfsResource:
            return
        if idx == 0:
            self.pinPath(self.currentIpfsResource, recursive=False)
        if idx == 1:
            self.pinPath(self.currentIpfsResource, recursive=True)

    def loadFromClipboardButtonClicked(self):
        current = self.app.clipTracker.getCurrent()
        if current:
            self.browseFsPath(current['path'])

    def onLoadIpfsCID(self):
        def onValidated(d):
            self.browseIpfsHash(d.getHash())

        runDialog(IPFSCIDInputDialog, title=iEnterIpfsCIDDialog(),
            accepted=onValidated)

    def onLoadIpfsMultipleCID(self):
        def onValidated(dlg):
            # Open a tab for every CID
            cids = dlg.getCIDs()
            for cid in cids:
                self.gWindow.addBrowserTab().browseIpfsHash(cid)

        runDialog(IPFSMultipleCIDInputDialog,
            title=iEnterIpfsCIDDialog(),
            accepted=onValidated)

    def onLoadIpns(self):
        text, ok = QInputDialog.getText(self,
                iEnterIpnsDialog(),
                iEnterIpns())
        if ok:
            self.browseIpnsHash(text)

    def onFollowIpns(self):
        runDialog(AddFeedDialog, self.app.marksLocal,
            self.currentIpfsResource,
            title=iFollowIpnsDialog())

    def onLoadHome(self):
        self.loadHomePage()

    def loadHomePage(self):
        homeUrl = self.app.settingsMgr.getSetting(CFG_SECTION_BROWSER,
            CFG_KEY_HOMEURL)
        self.enterUrl(QUrl(homeUrl))

    def refreshButtonClicked(self):
        self.ui.webEngineView.reload()

    def stopButtonClicked(self):
        self.ui.webEngineView.stop()
        self.ui.progressBar.setValue(0)

    def backButtonClicked(self):
        currentPage = self.ui.webEngineView.page()
        currentPage.history().back()

    def forwardButtonClicked(self):
        currentPage = self.ui.webEngineView.page()
        currentPage.history().forward()

    def onUrlChanged(self, url):
        if url.authority() == self.gatewayAuthority:
            self.currentIpfsResource = url.path()
            self.ui.urlZone.clear()
            # Content loaded from IPFS gateway, this is IPFS content
            self.ui.urlZone.insert(fsPath(url.path()))
            self.ipfsPathVisited.emit(self.currentIpfsResource)

            # Activate the follow action if this is IPNS
            # todo: better check on ipns path validity
            if self.currentIpfsResource.startswith('/ipns/'):
                self.followIpnsAction.setEnabled(True)
            else:
                self.followIpnsAction.setEnabled(False)
        else:
            self.currentIpfsResource = None
            self.ui.urlZone.clear()
            self.ui.urlZone.insert(url.toString())

    def onLoadFinished(self, ok):
        lenMax = 16
        pageTitle = self.ui.webEngineView.page().title()

        if pageTitle.startswith(self.gatewayAuthority):
            pageTitle = iNoTitle()

        self.currentPageTitle = pageTitle

        if len(pageTitle) > lenMax:
            pageTitle = '{0} ...'.format(pageTitle[0:lenMax])

        idx = self.gWindow.ui.tabWidget.indexOf(self)
        self.gWindow.ui.tabWidget.setTabText(idx, pageTitle)

        self.gWindow.ui.tabWidget.setTabToolTip(idx,
                self.currentPageTitle)

    def onIconChanged(self, icon):
        self.gWindow.ui.tabWidget.setTabIcon(self.tabPageIdx, icon)

    def onLoadProgress(self, progress):
        self.ui.progressBar.setValue(progress)

    def browseFsPath(self, path):
        self.enterUrl(QUrl('{0}:{1}'.format(SCHEME_IPFS, path)))

    def browseIpfsHash(self, ipfsHash):
        if not cidhelpers.cidValid(ipfsHash):
            return messageBox(iInvalidCID(ipfsHash))

        self.browseFsPath(joinIpfs(ipfsHash))

    def browseIpnsHash(self, ipnsHash):
        self.browseFsPath(joinIpns(ipnsHash))

    def enterUrl(self, url):
        self.ui.urlZone.clear()
        self.ui.urlZone.insert(url.toString())
        self.ui.webEngineView.load(url)

    def onUrlEdit(self):
        inputStr = self.ui.urlZone.text().strip()

        if cidhelpers.cidValid(inputStr):
            # Raw CID in the URL zone
            return self.browseIpfsHash(inputStr)

        url = QUrl(inputStr)
        scheme = url.scheme()

        if scheme in [SCHEME_FS, SCHEME_IPFS, SCHEME_DWEB]:
            self.enterUrl(url)
        elif scheme in ['http', 'https'] and \
                self.app.settingsMgr.allowHttpBrowsing is True:
            # Browse http urls if allowed
            self.enterUrl(url)
        else:
            messageBox(iInvalidUrl(inputStr))




import sys
import json

import aioipfs
import asyncio
import datetime

from galacteek import log as logger

from galacteek.ipfs.pubsub.messages import (MarksBroadcastMessage,
        PeerIdentMessageV1, PeerLogoutMessage)
from galacteek.ipfs.wrappers import *
from galacteek.ipfs.ipfsops import IPFSOperator
from galacteek.core.asynclib import asyncify

class PubsubService(object):
    """
    Generic IPFS pubsub service
    """

    def __init__(self, ipfsCtx, client, topic='galacteek.default',
            runPeriodic=False):
        self.client = client
        self.ipfsCtx = ipfsCtx
        self.topic = topic
        self.inQueue = asyncio.Queue()
        self.errorsQueue = asyncio.Queue()
        self.lock = asyncio.Lock()

        self._receivedCount = 0
        self._errorsCount = 0
        self._runPeriodic = runPeriodic

        self.tskServe = None
        self.tskProcess = None
        self.tskPeriodic = None

    @property
    def receivedCount(self):
        return self._receivedCount

    @property
    def errorsCount(self):
        return self._errorsCount

    @property
    def runPeriodic(self):
        return self._runPeriodic

    def debug(self, msg):
        logger.debug('PS[{0}]: {1}'.format(self.topic, msg))

    def logStatus(self):
        self.debug('** Messages received: {0}, errors: {1}'.format(
             self.receivedCount, self.errorsCount))

    def start(self):
        """ Create the different tasks for this service """
        self.tskServe = self.ipfsCtx.loop.create_task(self.serve())
        self.tskProcess = self.ipfsCtx.loop.create_task(self.processMessages())

        if self.runPeriodic:
            self.tskPeriodic = self.ipfsCtx.loop.create_task(self.periodic())

    async def stop(self):
        await self.shutdown()
        for tsk in [ self.tskServe, self.tskProcess, self.tskPeriodic ]:
            if not tsk:
                continue

            tsk.cancel()
            try:
                await tsk
            except asyncio.CancelledError as err:
                continue
            else:
                self.debug('task {}: shutdown ok'.format(tsk))

    async def shutdown(self):
        pass

    async def processMessages(self):
        """ Messages processing task to be implemented in your listener """
        return True

    async def periodic(self):
        return True

    async def serve(self):
        """
        Subscribe to the pubsub topic, read and filter incoming messages
        (dropping the ones emitted by us). Selected messages are put on the
        async queue (inQueue) to be later treated in the processMessages() task
        """

        nodeId = self.ipfsCtx.node.id

        try:
            async for message in self.client.pubsub.sub(self.topic):
                if message['from'].decode() == nodeId:
                    continue

                self.ipfsCtx.pubsub.psMessageRx.emit()

                await self.inQueue.put(message)
                self._receivedCount += 1
                await asyncio.sleep(0)
        except asyncio.CancelledError as e:
            self.debug('Cancelled, queue size was {0}'.format(
                self.inQueue.qsize()))
            return
        except Exception as e:
            self.debug('Serve interrupted by unknown exception {}'.format(
                str(e)))

    async def send(self, data, topic=None):
        """
        Publish a message

        :param str data: message payload
        """

        if topic is None:
            topic = self.topic

        status = await self.client.pubsub.pub(topic, data)
        self.ipfsCtx.pubsub.psMessageTx.emit()
        return status

class JSONPubsubService(PubsubService):
    """
    JSON pubsub listener, handling incoming messages as JSON objects
    """

    def msgDataToJson(self, msg):
        """
        Decode JSON data contained in a pubsub message
        """
        try:
            return json.loads(msg['data'].decode())
        except Exception as e:
            logger.debug('Could not decode JSON message data')
            return None

    async def processMessages(self):
        try:
            while True:
                data = await self.inQueue.get()

                if data is None:
                    break

                msg = self.msgDataToJson(data)
                if msg is None:
                    continue

                logger.debug('Received Pubsub message {}'.format(
                    json.dumps(msg, indent=4)))

                try:
                    await self.processJsonMessage(data['from'].decode(), msg)
                except Exception as exc:
                    logger.debug(
                            'processJsonMessage error: {}'.format(str(exc)))
                    await self.errorsQueue.put((msg, exc))
                    self._errorsCount += 1
        except asyncio.CancelledError as e:
            return
        except Exception as e:
            return

    async def processJsonMessage(self, sender, msg):
        """ Implement this method to process incoming JSON messages"""
        return True

class PSHashmarksExchanger(JSONPubsubService):
    def __init__(self, ipfsCtx, client, marksLocal, marksNetwork):
        super().__init__(ipfsCtx, client, topic='galacteek.ipfsmarks')

        self.marksLocal = marksLocal
        self.marksNetwork = marksNetwork
        self.marksLocal.markAdded.connect(self.onMarkAdded)

    @asyncify
    async def onMarkAdded(self, path, mark):
        if mark['share'] is True:
            await self.broadcastMarks({path: mark})

    async def broadcastMarks(self, marks):
        msg = MarksBroadcastMessage.make(marks)
        await self.send(str(msg))

    async def broadcastAllSharedMarks(self):
        all = self.marksLocal.getAll(share=True)
        if len(all.keys()) == 0:
            return
        msg = MarksBroadcastMessage.make(all)
        await self.send(str(msg))

    async def processJsonMessage(self, sender, msg):
        msgType = msg.get('msgtype', None)
        if msgType == MarksBroadcastMessage.TYPE:
            await self.processBroadcast(msg)

    async def processBroadcast(self, msg):
        marks = msg.get('marks', None)
        if not marks:
            return

        addedCount = 0
        for mark in marks.items():
            await asyncio.sleep(0)

            mPath = mark[0]
            if self.marksNetwork.search(mPath):
                continue

            category = 'auto'
            tsCreated = mark[1].get('tscreated', None)

            if tsCreated:
                date = datetime.datetime.fromtimestamp(tsCreated)
                category = '{0}/{1}/{2}'.format(date.year, date.month,
                        date.day)

            self.marksNetwork.insertMark(mark, category)
            addedCount += 1

        if addedCount > 0:
            with await self.lock:
                await self.marksNetwork.saveAsync()

            self.ipfsCtx.pubsubMarksReceived.emit(addedCount)

class PSMainService(JSONPubsubService):
    def __init__(self, ipfsCtx, client):
        super().__init__(ipfsCtx, client, topic='galacteek.main')

class PSPeersService(JSONPubsubService):
    def __init__(self, ipfsCtx, client):
        super().__init__(ipfsCtx, client, topic='galacteek.peers',
                runPeriodic=True)

        self._curProfile = None
        self._identEvery = 45
        self.ipfsCtx.profileChanged.connect(self.onProfileChanged)

    @property
    def curProfile(self):
        return self.ipfsCtx.currentProfile

    @property
    def identEvery(self):
        return self._identEvery

    @asyncify
    async def onProfileChanged(self, pName, profile):
        await profile.userInfo.loaded

        try:
            profile.userInfo.entryChanged.disconnect(self.userInfoChanged)
            profile.userInfo.available.disconnect(self.userInfoAvail)
        except Exception as e:
            pass

        self.curProfile.userInfo.entryChanged.connect(self.userInfoChanged)
        self.curProfile.userInfo.available.connect(self.userInfoAvail)
        await self.sendIdent(self.curProfile)

    @asyncify
    async def userInfoAvail(self, arg):
        await self.sendIdent(self.curProfile)

    @asyncify
    async def userInfoChanged(self):
        with await self.lock:
            await self.sendIdent(self.curProfile)

    @ipfsOp
    async def sendIdent(self, op, profile):
        nodeId = op.ctx.node.id
        uInfo = profile.userInfo

        if not uInfo.valid():
            return

        if uInfo.schemaVersion is 1:
            msg = PeerIdentMessageV1.make(nodeId, uInfo.objHash,
                    uInfo.root, profile.dagUser.dagCid,
                    profile.keyRootId, self.ipfsCtx.p2p.servicesFormatted())

            await self.send(str(msg))

    async def processJsonMessage(self, sender, msg):
        msgType = msg.get('msgtype', None)

        if msgType == PeerIdentMessageV1.TYPE:
            await self.handleIdentMessageV1(sender, msg)
        elif msgType == PeerLogoutMessage.TYPE:
            await self.handleLogoutMessage(sender, msg)

        await asyncio.sleep(0)

    async def handleLogoutMessage(self, sender, msg):
        lMsg = PeerLogoutMessage(msg)
        await self.ipfsCtx.peers.unregister(sender)

    async def handleIdentMessageV1(self, sender, msg):
        iMsg = PeerIdentMessageV1(msg)
        if not iMsg.valid():
            return

        if sender != iMsg.peer:
            # You forging pubsub messages, son ?
            return

        await self.ipfsCtx.peers.registerFromIdent(iMsg)

    @ipfsOp
    async def shutdown(self, op):
        msg = PeerLogoutMessage.make(op.ctx.node.id)
        await self.send(str(msg))

    async def periodic(self):
        while True:
            await asyncio.sleep(self.identEvery)

            with await self.lock:
                if self.curProfile:
                    await self.sendIdent(self.curProfile)

__all__ = [
        'PubsubService',
        'PSHashmarksExchanger',
        'PSMainService',
        'PSPeersService'
]
